#!/usr/bin/python

try: import simplejson as json
except ImportError: import json
import couchdb
import httplib
import urllib
import common
import unittest

HOST = "localhost:5984"
SET_NAME = "test_suite_set_view"
NUM_PARTS = 4
NUM_DOCS = 40000
DDOC = {
    "_id": "_design/test",
    "language": "javascript",
    "views": {
        "mapview1": {
            "map": "function(doc) { emit(doc.integer, doc.string); }"
        },
        "redview1": {
            "map": "function(doc) { emit(doc.integer, doc.string); }",
            "reduce": "function(keys, values, rereduce) {" + \
                "if (rereduce) {" + \
                "    return sum(values);" + \
                "} else {" + \
                "    return values.length;" + \
                "}" + \
             "}"
        },
        "redview2": {
            "map": "function(doc) { emit(doc.integer, doc.string); }",
            "reduce": "_count"
        }
    }
}


class TestPassivePartitions(unittest.TestCase):

    def setUp(self):
        self._params = {
            "host": HOST,
            "ddoc": DDOC,
            "nparts": NUM_PARTS,
            "ndocs": NUM_DOCS,
            "setname": SET_NAME,
            "server": couchdb.Server(url = "http://" + HOST)
            }
        # print "Creating databases"
        common.create_dbs(self._params)
        common.populate(self._params)
        common.define_set_view(self._params, range(NUM_PARTS), [])
        # print "Databases created"


    def tearDown(self):
        # print "Deleting test data"
        common.create_dbs(self._params, True)


    def test_passive_partitions(self):
        self.do_test_maps()
        self.do_test_reduce("redview1")
        self.do_test_reduce("redview2")
        self.do_test_view_updates()


    def do_test_maps(self):
        # print "Querying map view in steady state"
        (resp, view_result) = common.query(self._params, "mapview1")
        etag = resp.getheader("ETag")

        self.assertEqual(view_result["total_rows"], self._params["ndocs"],
                         "Query returned %d total_rows" % self._params["ndocs"])
        self.assertEqual(len(view_result["rows"]), self._params["ndocs"],
                         "Query returned %d rows" % self._params["ndocs"])

        common.test_keys_sorted(view_result)

        # print "Verifying set view group info"
        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2, 3], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        for i in [0, 1, 2, 3]:
            self.assertEqual(info["update_seqs"][str(i)], (self._params["ndocs"] / 4),
                             "right update seq for partition %d" % (i + 1))

        # print "Setting partition 4 as passive"
        common.set_partition_states(self._params, passive = [3])

        # print "Verifying set view group info"
        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [3], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        for i in [0, 1, 2, 3]:
            self.assertEqual(info["update_seqs"][str(i)], (self._params["ndocs"] / 4),
                             "right update seq for partition %d" % (i + 1))

        # print "Querying view again"
        (resp2, view_result2) = common.query(self._params, "mapview1")
        etag2 = resp2.getheader("ETag")

        expected = self._params["ndocs"] - (self._params["ndocs"] / 4)

        self.assertEqual(view_result2["total_rows"], self._params["ndocs"],
                         "Query returned %d total_rows" % self._params["ndocs"])
        self.assertEqual(len(view_result2["rows"]), expected,
                         "Query returned %d rows" % expected)
        self.assertNotEqual(etag2, etag, "Different Etag after setting passive partition")

        common.test_keys_sorted(view_result2)

        all_keys = {}
        for r in view_result2["rows"]:
            all_keys[r["key"]] = True

        for key in xrange(4, self._params["ndocs"], self._params["nparts"]):
            self.assertFalse(key in all_keys, "Key %d not in result after partition 4 was made passive" % key)

        # print "Setting partition 4 state to active"
        common.set_partition_states(self._params, active = [3])

        # print "Verifying set view group info"
        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2, 3], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        for i in [0, 1, 2, 3]:
            self.assertEqual(info["update_seqs"][str(i)], (self._params["ndocs"] / 4),
                             "right update seq for partition %d" % (i + 1))

        # print "Querying view again"
        (resp3, view_result3) = common.query(self._params, "mapview1")
        etag3 = resp3.getheader("ETag")

        self.assertEqual(view_result3["total_rows"], self._params["ndocs"],
                         "Query returned %d total_rows" % self._params["ndocs"])
        self.assertEqual(len(view_result3["rows"]), self._params["ndocs"],
                         "Query returned %d rows" % self._params["ndocs"])
        self.assertEqual(etag3, etag, "ETag is same as first query response")

        common.test_keys_sorted(view_result3)

        # print "Setting partitions 1 and 4 to passive state"
        common.set_partition_states(self._params, passive = [0, 3])

        # print "Verifying set view group info"
        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [1, 2], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [0, 3], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        for i in [0, 1, 2, 3]:
            self.assertEqual(info["update_seqs"][str(i)], (self._params["ndocs"] / 4),
                             "right update seq for partition %d" % (i + 1))

        # print "Querying view again"
        (resp4, view_result4) = common.query(self._params, "mapview1")
        etag4 = resp4.getheader("ETag")

        expected = self._params["ndocs"] / 2

        self.assertEqual(view_result4["total_rows"], self._params["ndocs"],
                         "Query returned %d total_rows" % self._params["ndocs"])
        self.assertEqual(len(view_result4["rows"]), expected,
                         "Query returned %d rows" % expected)
        self.assertNotEqual(etag4, etag, "ETag is different from all previous responses")
        self.assertNotEqual(etag4, etag2, "ETag is different from all previous responses")

        common.test_keys_sorted(view_result4)

        all_keys = {}
        for r in view_result4["rows"]:
            all_keys[r["key"]] = True

        for key in xrange(1, self._params["ndocs"], self._params["nparts"]):
            self.assertFalse(key in all_keys,
                             "Key %d not in result after partition 1 was made passive" % key)
        for key in xrange(4, self._params["ndocs"], self._params["nparts"]):
            self.assertFalse(key in all_keys,
                             "Key %d not in result after partition 4 was made passive" % key)

        # print "Marking all partitions as passive and querying view again"
        common.set_partition_states(self._params, passive = range(self._params["nparts"]))

        (resp5, view_result5) = common.query(self._params, "mapview1")
        etag5 = resp5.getheader("ETag")

        self.assertEqual(view_result5["total_rows"], self._params["ndocs"],
                         "Query returned %d total_rows" % self._params["ndocs"])
        self.assertEqual(len(view_result5["rows"]), 0, "Query returned 0 rows")
        self.assertNotEqual(etag5, etag, "ETag is different from all previous responses")
        self.assertNotEqual(etag5, etag2, "ETag is different from all previous responses")
        self.assertNotEqual(etag5, etag3, "ETag is different from all previous responses")
        self.assertNotEqual(etag5, etag4, "ETag is different from all previous responses")

        # print "Setting all partitions to active state"
        common.set_partition_states(self._params, active = range(self._params["nparts"]))

        # print "Verifying set view group info"
        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2, 3], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        for i in [0, 1, 2, 3]:
            self.assertEqual(info["update_seqs"][str(i)], (self._params["ndocs"] / 4),
                             "right update seq for partition %d" % (i + 1))

        # print "Querying view again"
        (resp6, view_result6) = common.query(self._params, "mapview1")
        etag6 = resp6.getheader("ETag")

        self.assertEqual(view_result6["total_rows"], self._params["ndocs"],
                         "Query returned %d total_rows" % self._params["ndocs"])
        self.assertEqual(len(view_result6["rows"]), self._params["ndocs"],
                         "Query returned %d rows" % self._params["ndocs"])
        self.assertEqual(etag6, etag, "ETag is the same from first view query response")

        common.test_keys_sorted(view_result6)


    def do_test_reduce(self, viewname):
        # print "Querying reduce view in steady state"
        (resp, view_result) = common.query(self._params, viewname)
        etag = resp.getheader("ETag")

        self.assertEqual(len(view_result["rows"]), 1, "Query returned 1 row")
        self.assertEqual(view_result["rows"][0]["value"], self._params["ndocs"],
                         "Non-grouped reduce value is %d" % self._params["ndocs"])

        # print "Verifying set view group info"
        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2, 3], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        for i in [0, 1, 2, 3]:
            self.assertEqual(info["update_seqs"][str(i)], (self._params["ndocs"] / 4),
                             "right update seq for partition %d" % (i + 1))

        # print "Setting partition 3 to passive state"
        common.set_partition_states(self._params, passive = [2])

        # print "Verifying set view group info"
        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 3], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [2], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        for i in [0, 1, 2, 3]:
            self.assertEqual(info["update_seqs"][str(i)], (self._params["ndocs"] / 4),
                             "right update seq for partition %d" % (i + 1))

        # print "Querying reduce view again"
        (resp2, view_result2) = common.query(self._params, viewname)
        etag2 = resp2.getheader("ETag")

        expected = self._params["ndocs"] - (self._params["ndocs"] / 4)
        self.assertEqual(len(view_result2["rows"]), 1, "Query returned 1 row")
        self.assertEqual(view_result2["rows"][0]["value"], expected,
                         "Non-grouped reduce value is %d" % expected)
        self.assertNotEqual(etag2, etag, "Different ETags")

        # print "Querying view with ?group=true"
        (resp3, view_result3) = common.query(self._params, viewname, {"group": "true"})
        etag3 = resp3.getheader("ETag")

        self.assertEqual(len(view_result3["rows"]), expected, "Query returned % rows" % expected)
        self.assertNotEqual(etag3, etag, "Different ETags")
        self.assertEqual(etag3, etag2, "Equal ETags for responses 2 and 3")

        common.test_keys_sorted(view_result3)

        all_keys = {}
        for r in view_result3["rows"]:
            all_keys[r["key"]] = True

        for key in xrange(3, self._params["ndocs"], self._params["nparts"]):
            self.assertFalse(key in all_keys,
                             "Key %d not in result after partition 3 was made passive" % key)

        # print "Querying view with ?group=true&descending=true"
        (resp4, view_result4) = common.query(
            self._params, viewname, {"group": "true", "descending": "true"})
        etag4 = resp4.getheader("ETag")

        self.assertEqual(len(view_result4["rows"]), expected, "Query returned % rows" % expected)
        self.assertNotEqual(etag4, etag, "Different ETags")
        self.assertEqual(etag4, etag3, "Equal ETags for responses 3 and 4")

        common.test_keys_sorted(view_result4, lambda a, b: a > b)

        all_keys = {}
        for r in view_result4["rows"]:
            all_keys[r["key"]] = True

        for key in xrange(3, self._params["ndocs"], self._params["nparts"]):
            self.assertFalse(key in all_keys,
                             "Key %d not in result after partition 3 was made passive" % key)

        # print "Querying view with ?group=true&startkey=3333&endkey=44781"
        (resp5, view_result5) = common.query(
            self._params, viewname,
            {"group": "true", "startkey": "3333", "endkey": "44781"})
        etag5 = resp5.getheader("ETag")

        self.assertNotEqual(etag5, etag, "Different ETags")
        self.assertEqual(etag5, etag4, "Equal ETags for responses 4 and 5")

        common.test_keys_sorted(view_result5)
        self.assertTrue(view_result5["rows"][0]["key"] >= 3333, "First key is >= 3333")
        self.assertTrue(view_result5["rows"][-1]["key"] <= 44781, "Last key is <= 44781")

        all_keys = {}
        for r in view_result5["rows"]:
            all_keys[r["key"]] = True

        for key in xrange(3, self._params["ndocs"], self._params["nparts"]):
            self.assertFalse(key in all_keys,
                             "Key %d not in result after partition 3 was made passive" % key)

        # print "Querying view with ?group=true&startkey=44781&endkey=3333&descending=true"
        (resp6, view_result6) = common.query(
            self._params, viewname,
            {"group": "true", "startkey": "44781", "endkey": "3333", "descending": "true"})
        etag6 = resp6.getheader("ETag")

        self.assertNotEqual(etag6, etag, "Different ETags")
        self.assertEqual(etag6, etag5, "Equal ETags for responses 5 and 6")

        common.test_keys_sorted(view_result6, lambda a, b: a > b)
        self.assertTrue(view_result6["rows"][0]["key"] <= 44781, "First key is <= 44781")
        self.assertTrue(view_result6["rows"][-1]["key"] >= 3333, "Last key is >= 3333")

        self.assertEqual(len(view_result6["rows"]), len(view_result5["rows"]),
                         "Same number of rows for responses 5 and 6")

        all_keys = {}
        for r in view_result6["rows"]:
            all_keys[r["key"]] = True

        for key in xrange(3, self._params["ndocs"], self._params["nparts"]):
            self.assertFalse(key in all_keys,
                             "Key %d not in result after partition 3 was made passive" % key)

        # print "Setting partition 3 to active state"
        common.set_partition_states(self._params, active = [2])

        # print "Verifying set view group info"
        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2, 3], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        for i in [0, 1, 2, 3]:
            self.assertEqual(info["update_seqs"][str(i)], (self._params["ndocs"] / 4),
                             "right update seq for partition %d" % (i + 1))

        # print "Querying view with ?group=true"
        (resp7, view_result7) = common.query(self._params, viewname, {"group": "true"})
        etag7 = resp7.getheader("ETag")

        self.assertEqual(len(view_result7["rows"]), self._params["ndocs"],
                         "Query returned % rows" % self._params["ndocs"])
        self.assertEqual(etag7, etag, "Same etags for responses 1 and 7")

        common.test_keys_sorted(view_result7)

        # print "Querying view with ?group=true&descending=true"
        (resp8, view_result8) = common.query(
            self._params, viewname, {"group": "true", "descending": "true"})
        etag8 = resp8.getheader("ETag")

        self.assertEqual(len(view_result8["rows"]), self._params["ndocs"],
                         "Query returned % rows" % self._params["ndocs"])
        self.assertEqual(etag7, etag8, "Same etags for responses 7 and 8")

        common.test_keys_sorted(view_result8, lambda a, b: a > b)

        # print "Querying view with ?group=true&startkey=3333&endkey=44781"
        (resp9, view_result9) = common.query(
            self._params, viewname,
            {"group": "true", "startkey": "3333", "endkey": "44781"})
        etag9 = resp9.getheader("ETag")

        self.assertEqual(etag9, etag8, "Equal ETags for responses 8 and 9")

        common.test_keys_sorted(view_result9)
        self.assertTrue(view_result9["rows"][0]["key"] >= 3333, "First key is >= 3333")
        self.assertTrue(view_result9["rows"][-1]["key"] <= 44781, "Last key is <= 44781")

        # print "Querying view with ?group=true&startkey=44781&endkey=3333&descending=true"
        (resp10, view_result10) = common.query(
            self._params, viewname,
            {"group": "true", "startkey": "44781", "endkey": "3333", "descending": "true"})
        etag10 = resp10.getheader("ETag")

        self.assertEqual(etag10, etag9, "Equal ETags for responses 9 and 10")

        common.test_keys_sorted(view_result10, lambda a, b: a > b)
        self.assertTrue(view_result10["rows"][0]["key"] <= 44781, "First key is <= 44781")
        self.assertTrue(view_result10["rows"][-1]["key"] >= 3333, "Last key is >= 3333")

        self.assertEqual(len(view_result10["rows"]), len(view_result9["rows"]),
                         "Same number of rows for responses 9 and 10")


    def do_test_view_updates(self):
        # print "Setting partition 2 state to passive"
        common.set_partition_states(self._params, passive = [1])

        # print "Verifying set view group info"
        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 2, 3], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [1], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        for i in [0, 1, 2, 3]:
            self.assertEqual(info["update_seqs"][str(i)], (self._params["ndocs"] / 4),
                             "right update seq for partition %d" % (i + 1))

        # print "Adding 2 new documents to partition 2"
        server = self._params["server"]
        db2 = server[self._params["setname"] + "/1"]
        new_doc1 = {"_id": "999999999", "integer": 999999999, "string": "999999999"}
        new_doc2 = {"_id": "000", "integer": -1111, "string": "000"}
        db2.save(new_doc1)
        db2.save(new_doc2)

        # print "Querying map view"
        (resp, view_result) = common.query(self._params, "mapview1")
        etag = resp.getheader("ETag")

        expected = self._params["ndocs"] - (self._params["ndocs"] / 4)

        self.assertEqual(len(view_result["rows"]), expected, "Query returned %d rows" % expected)

        common.test_keys_sorted(view_result)

        all_keys = {}
        for r in view_result["rows"]:
            all_keys[r["key"]] = True

        for key in xrange(2, self._params["ndocs"], self._params["nparts"]):
            self.assertFalse(key in all_keys,
                             "Key %d not in result after partition 2 was made passive" % key)
        self.assertFalse(new_doc1["integer"] in all_keys, "new_doc1 not reflected in view")
        self.assertFalse(new_doc2["integer"] in all_keys, "new_doc2 not reflected in view")

        # print "Setting partition 2 state to active"
        common.set_partition_states(self._params, active = [1])

        # print "Querying map view again"
        (resp2, view_result2) = common.query(self._params, "mapview1")
        etag2 = resp2.getheader("ETag")

        expected = self._params["ndocs"] + 2

        self.assertEqual(view_result2["total_rows"], expected, "Query returned %d total_rows" % expected)
        self.assertEqual(len(view_result2["rows"]), expected, "Query returned %d rows" % expected)

        common.test_keys_sorted(view_result2)

        all_keys = {}
        for r in view_result2["rows"]:
            all_keys[r["key"]] = True

        for key in xrange(2, self._params["ndocs"], self._params["nparts"]):
            self.assertTrue(key in all_keys,
                            "Key %d in result after partition 2 was re-enabled" % key)
        self.assertTrue(new_doc1["integer"] in all_keys, "new_doc1 reflected in view")
        self.assertTrue(new_doc2["integer"] in all_keys, "new_doc2 reflected in view")

        # print "Verifying set view group info"
        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2, 3], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        for i in [0, 1, 2, 3]:
            if i == 1:
                seq = (self._params["ndocs"] / 4) + 2
            else:
                seq = (self._params["ndocs"] / 4)
            self.assertEqual(info["update_seqs"][str(i)], seq, "right update seq for partition %d" % (i + 1))




""" Extend Queue class for python 2.4"""
import threading
import sys
from Queue import Queue

if sys.version >= "2.5":
    class PumpQueue(Queue):
        def __init__(self, maxsize=0):
            Queue.__init__(self, maxsize)
else:
    class PumpQueue(Queue):
        def __init__(self, maxsize=0):
            Queue.__init__(self, maxsize)
            self.all_tasks_done = threading.Condition(self.mutex)
            self.unfinished_tasks = 0

        def _put(self, item):
            Queue._put(self, item)
            self.unfinished_tasks += 1

        def task_done(self):
            self.all_tasks_done.acquire()
            try:
                unfinished = self.unfinished_tasks - 1
                if unfinished <= 0:
                    if unfinished < 0:
                        raise ValueError('task_done() called too many times')
                    self.all_tasks_done.notifyAll()
                self.unfinished_tasks = unfinished
            finally:
                self.all_tasks_done.release()

        def join(self):
            self.all_tasks_done.acquire()
            try:
                while self.unfinished_tasks:
                    self.all_tasks_done.wait()
            finally:
                self.all_tasks_done.release()



import stats_buffer
import util_cli as util

class NodeList:
    def run(self, accessor):
        result = []
        for node, node_info in stats_buffer.nodes.iteritems():
            result.append({"ip": node, "port": node_info['port'], "version" :node_info['version'], "os": node_info['os'], "status" :node_info['status']})
        return result

class NumNodes:
    def run(self, accessor):
        result = []
        result.append(len(stats_buffer.nodes))
        return result

class NumDownNodes:
    def run(self, accessor):
        result = []
        result.append(len(filter(lambda (a, b): b["status"]=="down", stats_buffer.nodes.items())))
        return result

class NumWarmupNodes:
    def run(self, accessor):
        result = []
        result.append(len(filter(lambda (a, b): b["status"]=="warmup", stats_buffer.nodes.items())))
        return result

class NumFailOverNodes:
    def run(self, accessor):
        result = []
        result.append(len(filter(lambda (a, b): b["clusterMembership"]!="active", stats_buffer.nodes.items())))
        return result

class BucketList:
    def run(self, accessor):
        result = []
        for bucket in stats_buffer.bucket_info.keys():
            result.append({"name": bucket})

        return result

class NodeStorageStats:
    def run(self, accessor):
        result = []
        for node, values in stats_buffer.nodes.iteritems():
            if values["StorageInfo"].has_key("hdd"):
                result.append({"ip": values["host"],
                           "port": values["port"],
                           "type" : "hdd",
                           "free": util.size_label(values["StorageInfo"]["hdd"]["free"]),
                           "quotaTotal" : util.size_label(values["StorageInfo"]["hdd"]["quotaTotal"]),
                           "used" : util.size_label(values["StorageInfo"]["hdd"]["used"]),
                           "usedByData" : util.size_label(values["StorageInfo"]["hdd"]["usedByData"]),
                           "total" : util.size_label(values["StorageInfo"]["hdd"]["total"])})
            if values["StorageInfo"].has_key("ram"):
                result.append({"ip": values["host"],
                           "port": values["port"],
                           "type" : "ram",
                           "quotaTotal" : util.size_label(values["StorageInfo"]["ram"]["quotaTotal"]),
                           "used" : util.size_label(values["StorageInfo"]["ram"]["used"]),
                           "usedByData" : util.size_label(values["StorageInfo"]["ram"]["usedByData"]),
                           "total" : util.size_label(values["StorageInfo"]["ram"]["total"])})
        return result

class NodeSystemStats:
    def run(self, accessor):
        result = []
        for node, values in stats_buffer.nodes.iteritems():
            result.append({"ip": values["host"],
                           "port": values["port"],
                           "cpuUtilization" :util.pretty_float(values["systemStats"]["cpu_utilization_rate"]),
                           "swapTotal": util.size_label(values["systemStats"]["swap_total"]),
                           "swapUsed" : util.size_label(values["systemStats"]["swap_used"]),
                           "currentItems" : values["systemStats"]["currentItems"],
                           "currentItemsTotal" : values["systemStats"]["currentItemsTotal"],
                           "replicaCurrentItems" : values["systemStats"]["replicaCurrentItems"]})

        return result

class ConnectionTrend:
    def run(self, accessor):
        result = {}
        for bucket, stats_info in stats_buffer.buckets.iteritems():
            values = stats_info[accessor["scale"]][accessor["counter"]]
            timestamps = values["timestamp"]
            timestamps = [x - timestamps[0] for x in timestamps]
            nodeStats = values["nodeStats"]
            samplesCount = values["samplesCount"]
            trend = []
            for node, vals in nodeStats.iteritems():
                a, b = util.linreg(timestamps, vals)
                trend.append((node, a, vals[-1]))
            result[bucket] = trend
        return result

class CalcTrend:
    def run(self, accessor):
        result = {}
        for bucket, stats_info in stats_buffer.buckets.iteritems():
            values = stats_info[accessor["scale"]][accessor["counter"]]
            timestamps = values["timestamp"]
            timestamps = [x - timestamps[0] for x in timestamps]
            nodeStats = values["nodeStats"]
            samplesCount = values["samplesCount"]
            trend = []
            for node, vals in nodeStats.iteritems():
                a, b = util.linreg(timestamps, vals)
                trend.append((node, a))
            result[bucket] = trend
        return result

class NodePerformanceStats:
    def run(self, accessor):
        result = {}
        for bucket, bucket_stats in stats_buffer.node_stats.iteritems():
            stats = []
            for node, stats_info in bucket_stats.iteritems():

                for key, value in stats_info.iteritems():
                    if key.find(accessor["counter"]) >= 0:
                        if accessor.has_key("threshold"):
                            if int(value) > accessor["threshold"]:
                                stats.append((node, (key, value)))
                        else:
                            if accessor.has_key("unit"):
                                if accessor["unit"] == "time":
                                    stats.append((node, util.time_label(value)))
                                elif accessor["unit"] == "size":
                                    stats.append((node, util.size_label(int(value))))
                            else:
                                stats.append((node, (key,value)))

            result[bucket] = stats
        return result

NodeCapsule = [
    {"name" : "NodeStatus",
     "ingredients" : [
        {
            "name" : "nodeList",
            "description" : "Node list",
            "code" : "NodeList",
        },
        {
            "name" : "numNodes",
            "description" : "Number of Nodes",
            "code" : "NumNodes",
        },
        {
            "name" : "numDownNodes",
            "description" : "Number of Down Nodes",
            "code" : "NumDownNodes",
        },
        {
            "name" : "numWarmupNodes",
            "description" : "Number of Warmup Nodes",
            "code" : "NumWarmupNodes",
        },
        {
            "name" : "numFailedOverNodes",
            "description" : "Number of Nodes failed over",
            "code" : "NumFailOverNodes",
        },
      ],
      "clusterwise" : False,
      "nodewise" : True,
      "perNode" : False,
      "perBucket" : False,
    },
    {"name" : "NumberOfConnection",
    "ingredients" : [
        {
            "name" : "connectionTrend",
            "description" : "Connection Trend",
            "counter" : "curr_connections",
            "scale" : "minute",
            "code" : "ConnectionTrend",
            "threshold" : {
                "high" : 10000,
            },
        },
     ],
     "nodewise" : True,
     "perNode" : True,
    },
    {"name" : "OOMError",
     "ingredients" : [
        {
            "name" : "oomErrors",
            "description" : "OOM Errors",
            "counter" : "ep_oom_errors",
            "scale" : "hour",
            "code" : "CalcTrend",
        },
        {
            "name" : "tempOomErrors",
            "description" : "Temporary OOM Errors",
            "counter" : "ep_tmp_oom_errors",
            "scale" : "hour",
            "code" : "CalcTrend",
        },
     ]
    },
    {"name" : "bucketList",
     "ingredients" : [
        {
            "name" : "bucketList",
            "description" : "Bucket list",
            "code" : "BucketList",
        },
     ],
     "nodewise" : True,
    },
    {"name" : "nodeStorageStats",
     "ingredients" : [
        {
            "name" : "nodeStorageStats",
            "description" : "Node storage stats",
            "code" : "NodeStorageStats",
        },
     ],
     "nodewise" : True,
    },
    {"name" : "nodeSystemStats",
     "ingredients" : [
        {
            "name" : "nodeSystemStats",
            "description" : "Node system stats",
            "code" : "NodeSystemStats",
        },
     ],
     "nodewise" : True,
    },
    {"name" : "tapPerformance",
     "ingredients" : [
        {
            "name" : "backfillRemaining",
            "description" : "Number of backfill remaining",
            "counter" : "ep_tap_queue_backfillremaining",
            "code" : "NodePerformanceStats",
        },
        {
            "name" : "tapNack",
            "description" : "Number of nacks",
            "counter" : "num_tap_nack",
            "code" : "NodePerformanceStats",
        },
        {
            "name" : "tapIdle",
            "description" : "Idle tap streams",
            "counter" : "idle",
            "code" : "NodePerformanceStats",
        },
     ],
     "perBucket" : True,
    },
    {"name" : "checkpointPerformance",
     "ingredients" : [
        {
            "name" : "openCheckPoint",
            "description" : "Items for open checkpoints",
            "counter" : "num_checkpoint_items",
            "code" : "NodePerformanceStats",
            "threshold" : 10000,
        },
     ],
     "perBucket" : True,
    },
    {"name" : "diskPerformance",
     "ingredients" : [
        {
            "name" : "diskCommit",
            "description" : "Averge disk commit time",
            "counter" : "disk_commit",
            "code" : "NodePerformanceStats",
            "unit" : "time",
        },
        {
            "name" : "diskUpdate",
            "description" : "Averge disk update time",
            "counter" : "disk_update",
            "code" : "NodePerformanceStats",
            "unit" : "time",
        },
        {
            "name" : "diskInsert",
            "description" : "Averge disk insert time",
            "counter" : "disk_insert",
            "code" : "NodePerformanceStats",
            "unit" : "time",
        },
        {
            "name" : "diskDelete",
            "description" : "Averge disk delete time",
            "counter" : "disk_del",
            "code" : "NodePerformanceStats",
            "unit" : "time",
        },
     ],
     "perBucket" : True,
    },
    {"name" : "AverageDocumentSize",
     "ingredients" : [
        {
            "name" : "averageDocumentSize",
            "description" : "Average Document Size",
            "counter" : "item_alloc_sizes",
            "code" : "NodePerformanceStats",
            "unit" : "size",
        },
     ],
     "perBucket" : True,
    },
    {"name" : "MemoryUsage",
     "ingredients" : [
        {
            "name" : "totalMemoryUsage",
            "description" : "Total memory usage",
            "counter" : "total_heap_bytes",
            "code" : "NodePerformanceStats",
            "unit" : "size",
        },
        {
            "name" : "totalFragmentation",
            "description" : "Total memory fragmentation",
            "counter" : "total_fragmentation_bytes",
            "code" : "NodePerformanceStats",
            "unit" : "size",
        },
        {
            "name" : "totalInternalMemory",
            "description" : "Total internal memory usage",
            "counter" : "mem_used",
            "code" : "NodePerformanceStats",
            "unit" : "size",
        },
        {
            "name" : "overhead",
            "description" : "Memory overhead",
            "counter" : "ep_overhead",
            "scale" : "hour",
            "code" : "NodePerformanceStats",
            "unit" : "size",
        },
     ],
     "perBucket" : True,
    },
    {"name" : "EPEnginePerformance",
     "ingredients" : [
        {
            "name" : "flusherState",
            "description" : "Engine flusher state",
            "counter" : "ep_flusher_state",
            "code" : "NodePerformanceStats",
        },
        {
            "name" : "flusherCompleted",
            "description" : "Flusher completed",
            "counter" : "ep_flusher_num_completed",
            "code" : "NodePerformanceStats",
        },
        {
            "name" : "avgItemLoadTime",
            "description" : "Average item loaded time",
            "counter" : "ep_bg_load_avg",
            "code" : "NodePerformanceStats",
            "unit" : "time"
        },
        {
            "name" : "avgItemWaitTime",
            "description" : "Averge item waited time",
            "counter" : "ep_bg_wait_avg",
            "code" : "NodePerformanceStats",
            "unit" : "time",
        },
     ],
     "perNode" : True,
    },
]





import sys
import string
import socket
import select
import unittest
import threading
import time
import re
import struct

from memcacheConstants import REQ_MAGIC_BYTE, RES_MAGIC_BYTE
from memcacheConstants import REQ_PKT_FMT, RES_PKT_FMT, MIN_RECV_PACKET
from memcacheConstants import SET_PKT_FMT, DEL_PKT_FMT, INCRDECR_RES_FMT

import memcacheConstants

import moxi_mock_server

# Before you run moxi_mock_b2b.py, start a moxi like...
#
#   ./moxi -z 11333=localhost:11311 -p 0 -U 0 -vvv -t 1 -O stderr
#          -Z downstream_max=1,downstream_conn_max=0
#
# Or, if you're using the vbucket-aware moxi...
#
#   ./moxi -z ./t/moxi_mock.cfg -p 0 -U 0 -vvv -t 1 -O stderr
#          -Z downstream_max=1,downstream_conn_max=0,downstream_protocol=binary
#
# Then...
#
#   python ./t/moxi_mock_b2b.py
#
# ----------------------------------

class TestProxyBinary(moxi_mock_server.ProxyClientBase):
    def __init__(self, x):
        moxi_mock_server.ProxyClientBase.__init__(self, x)

    def SOON_testBasicVersion(self):
        """Test version command does not reach mock server"""
        self.client_connect()
        self.client_send("version\r\n")
        self.client_recv("VERSION .*\r\n")
        self.assertTrue(self.mock_quiet())
        self.client_connect()
        self.client_send("version\r\n")
        self.client_recv("VERSION .*\r\n")
        self.assertTrue(self.mock_quiet())

    def testBasicSerialGet(self):
        """Test basic serial get commands"""
        self.client_connect()

        self.client_send(self.packReq(memcacheConstants.CMD_GETK, key='keyNotThere0'))
        self.mock_recv(self.packReq(memcacheConstants.CMD_GETK, key='keyNotThere0'))
        self.mock_send(self.packRes(memcacheConstants.CMD_GETK,
                                    status=memcacheConstants.ERR_NOT_FOUND,
                                    key='keyNotThere0'))
        self.client_recv(self.packRes(memcacheConstants.CMD_GETK,
                                    status=memcacheConstants.ERR_NOT_FOUND,
                                    key='keyNotThere0'))

        self.client_send(self.packReq(memcacheConstants.CMD_GETK, key='keyNotThere0'))
        self.mock_recv(self.packReq(memcacheConstants.CMD_GETK, key='keyNotThere0'))
        self.mock_send(self.packRes(memcacheConstants.CMD_GETK,
                                    status=memcacheConstants.ERR_NOT_FOUND,
                                    key='keyNotThere0'))
        self.client_recv(self.packRes(memcacheConstants.CMD_GETK,
                                    status=memcacheConstants.ERR_NOT_FOUND,
                                    key='keyNotThere0'))

        self.client_send(self.packReq(memcacheConstants.CMD_GETK, key='keyNotThere0'))
        self.mock_recv(self.packReq(memcacheConstants.CMD_GETK, key='keyNotThere0'))
        self.mock_send(self.packRes(memcacheConstants.CMD_GETK,
                                    status=memcacheConstants.ERR_NOT_FOUND,
                                    key='keyNotThere0'))
        self.client_recv(self.packRes(memcacheConstants.CMD_GETK,
                                    status=memcacheConstants.ERR_NOT_FOUND,
                                    key='keyNotThere0'))

    def testBasicQuit(self):
        """Test quit command does not reach mock server"""
        self.client_connect()

        self.client_send(self.packReq(memcacheConstants.CMD_GETK, key='keyNotThere0'))
        self.mock_recv(self.packReq(memcacheConstants.CMD_GETK, key='keyNotThere0'))
        self.mock_send(self.packRes(memcacheConstants.CMD_GETK,
                                    status=memcacheConstants.ERR_NOT_FOUND,
                                    key='keyNotThere0'))
        self.client_recv(self.packRes(memcacheConstants.CMD_GETK,
                                    status=memcacheConstants.ERR_NOT_FOUND,
                                    key='keyNotThere0'))

        self.client_send(self.packReq(memcacheConstants.CMD_QUIT))
        self.assertTrue(self.mock_quiet())

    def SOON_testBogusCommand(self):
        """Test bogus commands do not reach mock server"""
        self.client_connect()
        self.client_send('bogus\r\n')
        self.client_recv('.*ERROR.*\r\n')
        self.assertTrue(self.mock_quiet())

    def doTestSimpleSet(self, flg=0, exp=0, val='1'):
        """Test simple set against mock server"""
        self.client_connect()
        self.client_send(self.packReq(memcacheConstants.CMD_SET, key='simpleSet',
                                      extraHeader=struct.pack(memcacheConstants.SET_PKT_FMT, flg, exp),
                                      val=val))
        self.mock_recv(self.packReq(memcacheConstants.CMD_SET, key='simpleSet',
                                    extraHeader=struct.pack(memcacheConstants.SET_PKT_FMT, flg, exp),
                                    val=val))
        self.mock_send(self.packRes(memcacheConstants.CMD_SET, status=0))
        self.client_recv(self.packRes(memcacheConstants.CMD_SET, status=0))

    def testSimpleSet(self):
        self.doTestSimpleSet()

    def testSimpleSetWithParams(self):
        self.doTestSimpleSet(flg=1234, exp=4321)

    def testSimpleSetWithEmptyVal(self):
        self.doTestSimpleSet(flg=1234, exp=4321, val='')

    def doTestFlushAllBroadcast(self, exp=0):
        """Test flush_all scatter/gather"""
        self.client_connect()
        self.client_send(self.packReq(memcacheConstants.CMD_FLUSH,
                                    extraHeader=struct.pack(memcacheConstants.FLUSH_PKT_FMT, exp)))
        self.mock_recv(self.packReq(memcacheConstants.CMD_FLUSH,
                                    extraHeader=struct.pack(memcacheConstants.FLUSH_PKT_FMT, exp)))
        self.mock_send(self.packRes(memcacheConstants.CMD_FLUSH, status=0))
        self.client_recv(self.packRes(memcacheConstants.CMD_FLUSH, status=0))

    def testFlushAllBroadcast(self):
        self.doTestFlushAllBroadcast()

    def testFlushAllWithTime(self):
        self.doTestFlushAllBroadcast(exp=1234)

    def testSplitResponseOverSeveralWrites(self):
        """Test split a response over several writes"""
        self.client_connect()
        self.client_send(self.packReq(memcacheConstants.CMD_SET, key='splitResponse',
                                      extraHeader=struct.pack(memcacheConstants.SET_PKT_FMT, 0, 0),
                                      val='1'))
        self.mock_recv(self.packReq(memcacheConstants.CMD_SET, key='splitResponse',
                                    extraHeader=struct.pack(memcacheConstants.SET_PKT_FMT, 0, 0),
                                    val='1'))
        r = self.packRes(memcacheConstants.CMD_SET, status=0)
        self.mock_send(r[0:3])
        self.wait(1)
        self.mock_send(r[3:6])
        self.wait(1)
        self.mock_send(r[6:22])
        self.wait(1)
        self.mock_send(r[22:]) # Header is 24 bytes.
        self.client_recv(r)

    def testSplitRequestOverSeveralWrites(self):
        """Test split a request over several writes"""
        self.client_connect()
        r = self.packReq(memcacheConstants.CMD_SET, key='splitRequest',
                         extraHeader=struct.pack(memcacheConstants.SET_PKT_FMT, 0, 0),
                         val='hello')
        self.client_send(r[0:3])
        self.wait(1)
        self.client_send(r[3:24])
        self.wait(1)
        self.client_send(r[24:26])
        self.wait(1)
        self.client_send(r[26:])

        self.mock_recv(r)
        self.mock_send(self.packRes(memcacheConstants.CMD_SET, status=0))
        self.client_recv(self.packRes(memcacheConstants.CMD_SET, status=0))

    def testTerminateResponseWithServerClose(self):
        """Test chop the response with a server close"""
        self.client_connect()
        r = self.packReq(memcacheConstants.CMD_SET, key='chopped',
                                    extraHeader=struct.pack(memcacheConstants.SET_PKT_FMT, 0, 0),
                                    val='1')
        self.client_send(r)
        self.mock_recv(r)
        self.mock_close()
        self.client_recv('.*ERROR .*\r\n') # Oddly, this passes due to the error message.

    def testMultiGetValueAllMiss(self):
        """Test the proxy handles VALUE response"""
        self.client_connect()
        r = (self.packReq(memcacheConstants.CMD_GETKQ, key='someVal0', opaque=4) +
             self.packReq(memcacheConstants.CMD_GETKQ, key='someVal1', opaque=13) +
             self.packReq(memcacheConstants.CMD_NOOP))
        self.client_send(r)
        self.mock_recv(r)
        self.mock_send(self.packRes(memcacheConstants.CMD_NOOP))
        self.client_recv(self.packRes(memcacheConstants.CMD_NOOP))

    def testMultiGetValueSomeHit(self):
        self.client_connect()
        r = (self.packReq(memcacheConstants.CMD_GETKQ, key='someVal0', opaque=4) +
             self.packReq(memcacheConstants.CMD_GETKQ, key='someVal1', opaque=13) +
             self.packReq(memcacheConstants.CMD_NOOP))
        self.client_send(r)
        self.mock_recv(r)
        self.mock_send(self.packRes(memcacheConstants.CMD_GETKQ, key='someVal0', opaque=4,
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 567),
                                    val='0123456789'))
        self.mock_send(self.packRes(memcacheConstants.CMD_NOOP))
        self.client_recv(self.packRes(memcacheConstants.CMD_GETKQ, key='someVal0', opaque=4,
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 567),
                                    val='0123456789') +
                         self.packRes(memcacheConstants.CMD_NOOP))

    def testMultiGetValueAllHit(self):
        self.client_connect()
        r = (self.packReq(memcacheConstants.CMD_GETKQ, key='someVal0', opaque=4) +
             self.packReq(memcacheConstants.CMD_GETKQ, key='someVal1', opaque=13) +
             self.packReq(memcacheConstants.CMD_NOOP))
        self.client_send(r)
        self.mock_recv(r)
        self.mock_send(self.packRes(memcacheConstants.CMD_GETKQ, key='someVal0', opaque=4,
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                                    val='0123456789'))
        self.mock_send(self.packRes(memcacheConstants.CMD_GETKQ, key='someVal1', opaque=13,
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                                    val='0123456789'))
        self.mock_send(self.packRes(memcacheConstants.CMD_NOOP))
        self.client_recv(
            self.packRes(memcacheConstants.CMD_GETKQ, key='someVal0', opaque=4,
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                                    val='0123456789') +
            self.packRes(memcacheConstants.CMD_GETKQ, key='someVal1', opaque=13,
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                                    val='0123456789') +
            self.packRes(memcacheConstants.CMD_NOOP))

    def testMultiGetValueAllReversedHit(self):
        self.client_connect()
        r = (self.packReq(memcacheConstants.CMD_GETKQ, key='someVal0', opaque=4) +
             self.packReq(memcacheConstants.CMD_GETKQ, key='someVal1', opaque=13) +
             self.packReq(memcacheConstants.CMD_NOOP))
        self.client_send(r)
        self.mock_recv(r)
        self.mock_send(self.packRes(memcacheConstants.CMD_GETKQ, key='someVal1', opaque=13,
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                                    val='0123456789'))
        self.mock_send(self.packRes(memcacheConstants.CMD_GETKQ, key='someVal0', opaque=4,
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                                    val='0123456789'))
        self.mock_send(self.packRes(memcacheConstants.CMD_NOOP))
        self.client_recv(
            self.packRes(memcacheConstants.CMD_GETKQ, key='someVal1', opaque=13,
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                                    val='0123456789') +
            self.packRes(memcacheConstants.CMD_GETKQ, key='someVal0', opaque=4,
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                                    val='0123456789') +
            self.packRes(memcacheConstants.CMD_NOOP))

    def testGetEmptyValue(self):
        """Test the proxy handles empty VALUE response"""
        self.client_connect()
        r = self.packReq(memcacheConstants.CMD_GETK, key='someVal')
        self.client_send(r)
        self.mock_recv(r)
        q = self.packRes(memcacheConstants.CMD_GETK, key='someVal',
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                                    val='')
        self.mock_send(q)
        self.client_recv(q)

    def SOON_testTerminateResponseWithServerCloseInValue(self):
        """Test chop the VALUE response with a server close"""
        self.client_connect()
        self.client_send('get someChoppedVal\r\n')
        self.mock_recv(self.packReq(memcacheConstants.CMD_GETK, key='someChoppedVal'))
        r = self.packRes(memcacheConstants.CMD_GETK, key='someChoppedVal',
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                                    val='0123456789')
        self.mock_send(r[0:30]) # Header (24) + Ext (4) + Value (10) == 38, so missing a few bytes.
        self.mock_close()
        self.client_recv('END\r\n')

    def SOON_testTerminateResponseWithServerCloseIn2ndValue(self):
        """Test chop the 2nd VALUE response with a server close"""
        self.client_connect()
        self.client_send('get someWholeVal someChoppedVal\r\n')
        self.mock_recv(self.packReq(memcacheConstants.CMD_GETKQ, key='someWholeVal', opaque=4) +
                       self.packReq(memcacheConstants.CMD_GETKQ, key='someChoppedVal', opaque=17) +
                       self.packReq(memcacheConstants.CMD_NOOP))
        r = (self.packRes(memcacheConstants.CMD_GETK, key='someWholeVal',
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                                    opaque=4,
                                    val='0123456789') +
             self.packRes(memcacheConstants.CMD_GETK, key='someChoppedVal',
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                                    opaque=17,
                                    val='0123456789') +
             self.packRes(memcacheConstants.CMD_NOOP))
        self.mock_send(r[0:(24 + 4 + 12 + 10 + 5)])
        self.mock_close()
        self.client_recv('VALUE someWholeVal 0 10\r\n0123456789\r\nEND\r\n')

    def SOON_testTerminateResponseWithServerCloseIn2ndValueData(self):
        """Test chop the 2nd VALUE data response with a server close"""
        self.client_connect()
        self.client_send('get someWholeVal someChoppedVal\r\n')
        self.mock_recv(self.packReq(memcacheConstants.CMD_GETKQ, key='someWholeVal', opaque=4) +
                       self.packReq(memcacheConstants.CMD_GETKQ, key='someChoppedVal', opaque=17) +
                       self.packReq(memcacheConstants.CMD_NOOP))
        r = (self.packRes(memcacheConstants.CMD_GETK, key='someWholeVal',
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                                    opaque=4,
                                    val='0123456789') +
             self.packRes(memcacheConstants.CMD_GETK, key='someChoppedVal',
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                                    opaque=17,
                                    val='0123456789') +
             self.packRes(memcacheConstants.CMD_NOOP))
        self.mock_send(r[0:(24 + 4 + 12 + 10 + 24 + 4 + 20)])
        self.mock_close()
        self.client_recv('VALUE someWholeVal 0 10\r\n0123456789\r\nEND\r\n')

    def SOON_testTerminateResponseWithServerCloseInNOOP(self):
        """Test chop NOOP terminator"""
        self.client_connect()
        self.client_send('get someWholeVal someChoppedVal\r\n')
        self.mock_recv(self.packReq(memcacheConstants.CMD_GETKQ, key='someWholeVal', opaque=4) +
                       self.packReq(memcacheConstants.CMD_GETKQ, key='someChoppedVal', opaque=17) +
                       self.packReq(memcacheConstants.CMD_NOOP))
        r = (self.packRes(memcacheConstants.CMD_GETK, key='someWholeVal',
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                                    opaque=4,
                                    val='0123456789') +
             self.packRes(memcacheConstants.CMD_GETK, key='someChoppedVal',
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                                    opaque=17,
                                    val='0123456789') +
             self.packRes(memcacheConstants.CMD_NOOP))
        self.mock_send(r[:-4])
        self.mock_close()
        # TODO: Need to diagnose why moxi gives this response.
        self.client_recv('VALUE someWholeVal 0 10\r\n0123456789\r\n' +
                         'END\r\n')

    def SOON_testTerminateResponseWithServerCloseAfterValueHeader(self):
        """Test chop response after 2nd item header"""
        self.client_connect()
        self.client_send('get someWholeVal someChoppedVal\r\n')
        self.mock_recv(self.packReq(memcacheConstants.CMD_GETKQ, key='someWholeVal', opaque=4) +
                       self.packReq(memcacheConstants.CMD_GETKQ, key='someChoppedVal', opaque=17) +
                       self.packReq(memcacheConstants.CMD_NOOP))
        r = (self.packRes(memcacheConstants.CMD_GETK, key='someWholeVal',
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                                    opaque=4,
                                    val='0123456789') +
             self.packRes(memcacheConstants.CMD_GETK, key='someChoppedVal',
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                                    opaque=17,
                                    val='0123456789') +
             self.packRes(memcacheConstants.CMD_NOOP))
        self.mock_send(r[0:(24 + 4 + 12 + 10 + 24)])
        self.mock_close()
        self.client_recv('VALUE someWholeVal 0 10\r\n0123456789\r\n' +
                         'END\r\n')

    def SOON_testTerminateResponseWithServerCloseAfter1stItem(self):
        """Test chop response after 1st item"""
        self.client_connect()
        self.client_send('get someWholeVal someChoppedVal\r\n')
        self.mock_recv(self.packReq(memcacheConstants.CMD_GETKQ, key='someWholeVal', opaque=4) +
                       self.packReq(memcacheConstants.CMD_GETKQ, key='someChoppedVal', opaque=17) +
                       self.packReq(memcacheConstants.CMD_NOOP))
        r = (self.packRes(memcacheConstants.CMD_GETK, key='someWholeVal',
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                                    opaque=4,
                                    val='0123456789') +
             self.packRes(memcacheConstants.CMD_GETK, key='someChoppedVal',
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                                    opaque=17,
                                    val='0123456789') +
             self.packRes(memcacheConstants.CMD_NOOP))
        self.mock_send(r[0:(24 + 4 + 12 + 10)])
        self.mock_close()
        self.client_recv('VALUE someWholeVal 0 10\r\n0123456789\r\n' +
                         'END\r\n')

    def testServerGoingDownAndUp(self):
        """Test server going up and down with no client impact"""
        self.client_connect()
        r = self.packReq(memcacheConstants.CMD_GETK, key='someUp')
        self.client_send(r)
        self.mock_recv(r)
        q = self.packRes(memcacheConstants.CMD_GETK, key='someUp',
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                                    val='0123456789')
        self.mock_send(q)
        self.client_recv(q)

        self.mock_close()

        self.client_send(r)
        self.mock_recv(r)
        self.mock_send(q)
        self.client_recv(q)

    def testServerGoingDownAndUpAfterResponse(self):
        """Test server going up and down after response with no client impact"""
        self.client_connect()
        r = self.packReq(memcacheConstants.CMD_GETK, key='someUp')
        self.client_send(r)
        self.mock_recv(r)
        q = self.packRes(memcacheConstants.CMD_GETK, key='someUp',
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                                    val='0123456789')
        self.mock_send(q)

        self.mock_close() # Try close before client_recv().

        self.client_recv(q)

        self.client_send(r)
        self.mock_recv(r)
        self.mock_send(q)
        self.client_recv(q)

    def testTwoSerialClients(self):
        """Test two serial clients"""

        # Assuming proxy's downstream_max is 1,
        # and number of threads is 1.

        self.client_connect(0)
        r = self.packReq(memcacheConstants.CMD_GETK, key='client0')
        self.client_send(r)
        self.mock_recv(r)
        q = self.packRes(memcacheConstants.CMD_GETK, key='client0',
                         extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                         val='0123456789')
        self.mock_send(q)
        self.client_recv(q)

        # Note that mock server sees 1 session that's reused
        # even though two clients are connected.

        self.client_connect(1)
        r = self.packReq(memcacheConstants.CMD_GETK, key='client1')
        self.client_send(r)
        self.mock_recv(r)
        q = self.packRes(memcacheConstants.CMD_GETK, key='client1',
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                                    val='0123456789')
        self.mock_send(q)
        self.client_recv(q)

    def testTwoSerialClientsConnectingUpfront(self):
        """Test two serial clients that both connect upfront"""

        # Assuming proxy's downstream_max is 1,
        # and number of threads is 1.

        self.client_connect(0)
        self.client_connect(1)

        r = self.packReq(memcacheConstants.CMD_GETK, key='client0')
        self.client_send(r)
        self.mock_recv(r)
        q = self.packRes(memcacheConstants.CMD_GETK, key='client0',
                         extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                         val='0123456789')
        self.mock_send(q)
        self.client_recv(q)

        # Note that mock server sees 1 session that's reused
        # even though two clients are connected.

        r = self.packReq(memcacheConstants.CMD_GETK, key='client1')
        self.client_send(r)
        self.mock_recv(r)
        q = self.packRes(memcacheConstants.CMD_GETK, key='client1',
                                    extraHeader=struct.pack(memcacheConstants.GET_RES_FMT, 0),
                                    val='0123456789')
        self.mock_send(q)
        self.client_recv(q)

    # Dedupe of keys is disabled for now in vbucket-aware moxi.
    #
    def TODO_testGetSquash(self):
        """Test multiget by multiple clients are deduped"""

        # Assuming proxy's downstream_max is 1,
        # and number of threads is 1.

        self.client_connect(0)
        self.client_connect(1)
        self.client_connect(2)

        self.client_send('get cork0\r\n', 0)
        self.mock_recv('get cork0\r\n', 0)

        # Mock server is 'busy' at this point, so
        # any client sends should be able to be
        # de-duplicated by the proxy.

        self.client_send('get a b c\r\n', 1)
        self.client_send('get b c d\r\n', 2)

        self.wait(10)

        self.mock_send('END\r\n', 0)
        self.client_recv('END\r\n', 0)

        self.mock_recv('get a b c d\r\n', 0)
        self.mock_send('VALUE a 0 1\r\na\r\n', 0)
        self.mock_send('VALUE b 0 1\r\nb\r\n', 0)
        self.mock_send('VALUE c 0 1\r\nc\r\n', 0)
        self.mock_send('VALUE d 0 1\r\nd\r\n', 0)
        self.mock_send('END\r\n', 0)

        self.client_recv('VALUE a 0 1\r\na\r\n' +
                         'VALUE b 0 1\r\nb\r\n' +
                         'VALUE c 0 1\r\nc\r\n' +
                         'END\r\n', 1)
        self.client_recv('VALUE b 0 1\r\nb\r\n' +
                         'VALUE c 0 1\r\nc\r\n' +
                         'VALUE d 0 1\r\nd\r\n' +
                         'END\r\n', 2)

    # Dedupe of keys is disabled for now in vbucket-aware moxi.
    #
    def TODO_testGetSquashOneKey(self):
        """Test multiget of one key by multiple clients are deduped"""

        # Assuming proxy's downstream_max is 1,
        # and number of threads is 1.

        self.client_connect(0)
        self.client_connect(1)
        self.client_connect(2)
        self.client_connect(3)
        self.client_connect(4)

        self.client_send('get wait0\r\n', 0)
        self.mock_recv('get wait0\r\n', 0)

        # Mock server is 'busy' at this point, so
        # any client sends should be able to be
        # de-duplicated by the proxy.

        self.client_send('get a\r\n', 1)
        self.client_send('get a\r\n', 2)
        self.client_send('get a\r\n', 3)
        self.client_send('get a\r\n', 4)

        self.wait(10)

        self.mock_send('END\r\n', 0)
        self.client_recv('END\r\n', 0)

        self.mock_recv('get a\r\n', 0)
        self.mock_send('VALUE a 0 1\r\na\r\n', 0)
        self.mock_send('END\r\n', 0)

        self.client_recv('VALUE a 0 1\r\na\r\n' +
                         'END\r\n', 1)
        self.client_recv('VALUE a 0 1\r\na\r\n' +
                         'END\r\n', 2)
        self.client_recv('VALUE a 0 1\r\na\r\n' +
                         'END\r\n', 3)
        self.client_recv('VALUE a 0 1\r\na\r\n' +
                         'END\r\n', 4)

    # Dedupe of keys is disabled for now in vbucket-aware moxi.
    #
    def TODO_testGetSquashNoKeyOverlap(self):
        """Test multiget dedupe, but no key overlap"""

        # Assuming proxy's downstream_max is 1,
        # and number of threads is 1.

        self.client_connect(0)
        self.client_connect(1)
        self.client_connect(2)

        self.client_send('get cork0\r\n', 0)
        self.mock_recv('get cork0\r\n', 0)

        # Mock server is 'busy' at this point, so
        # any client sends should be able to be
        # de-duplicated by the proxy.

        self.client_send('get a\r\n', 1)
        self.client_send('get x\r\n', 2)

        self.wait(10)

        self.mock_send('END\r\n', 0)
        self.client_recv('END\r\n', 0)

        self.mock_recv('get a x\r\n', 0)
        self.mock_send('VALUE a 0 1\r\na\r\n', 0)
        self.mock_send('VALUE x 0 1\r\nx\r\n', 0)
        self.mock_send('END\r\n', 0)

        self.client_recv('VALUE a 0 1\r\na\r\n' +
                         'END\r\n', 1)
        self.client_recv('VALUE x 0 1\r\nx\r\n' +
                         'END\r\n', 2)

    def TODO_testTimeout(self):
        """Test downstream timeout handling"""
        return """TODO: Highly dependent on hardcoded downstream timeout val"""

        # Assuming proxy's downstream_max is 1,
        # and number of threads is 1.

        self.client_connect(0)

        self.client_send('get time0\r\n', 0)
        self.mock_recv('get time0\r\n', 0)

        # Mock server is 'busy' at this point, so
        # downstream timeout logic should kick in,
        # without our mock server having to send anything.

        self.wait(210)

        self.client_recv('END\r\n', 0)

        # TODO: The number of server sessions should be 0,
        # except the close might not have propagated.

    def TODO_testSharedServerConns(self):
        """Test proxy only uses a few server conns"""
        return "TODO: getting nondetermistic behavior here due to retry feature"

        self.assertEqual(len(self.clients), 0)
        self.assertEqual(len(self.mock_server().sessions), 0)

        large_num_clients = 30
        for i in range(0, large_num_clients):
            self.client_connect(i)
            self.client_send('get fromClient' + str(i) + '\r\n', i)

        self.assertEqual(len(self.clients), large_num_clients)

        self.wait(10)

        self.assertTrue(len(self.mock_server().sessions) < large_num_clients)
        self.assertTrue(len(self.mock_server().sessions) > 0)

    def TODO_testServerSeesRetry(self):
        """Test server going down sees a retry"""
        return "TODO: getting nondetermistic behavior here due to retry feature"

        self.client_connect()
        self.client_send('get someDown\r\n')
        self.mock_recv("get someDown\r\n")
        self.mock_send('VALUE someDown 0 10\r\n')

        self.mock_close() # Go down before full reply, we should see retry.

        self.mock_recv("get someDown\r\n")
        self.mock_send('VALUE someDown 0 10\r\n')
        self.mock_send('0123456789\r\n')
        self.mock_send('END\r\n')
        self.client_recv('VALUE someDown 0 10\r\n0123456789\r\nEND\r\n')

# More test ideas...
#
# Test chopped up responses from multiple mock servers.
# Test chopped up requests from multiple clients.
# Test servers going down during multiget write.
# Test chopped up server responses during multiget.
# Test if one server goes down during a broadcast/scatter-gather.
# Test server going down.
# Test server coming back up.
# Test server getting very slow.
# Test several backend servers, and one or more start flapping.
## test get and multi-get
## test simple update commands
## test broadcast commands like flush_all
#
# Retest all the above with binary protocol.

if __name__ == '__main__':
    unittest.main()




#
# Copyright 2012, Couchbase, Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import json
import migrator

sources = [{'type':'json', 'class':'JSONReader',
            'example':'json://<filename>'}]
destinations = [{'type':'json', 'class':'JSONWriter',
                 'example':'json://<filename>'}]


class JSONReader(migrator.Reader):
    def __init__(self, source):
        if source[0:2] == "//":
            source = source[2:]
        self.reader = open(source, 'rb')

    def __iter__(self):
        return self

    def next(self):
        data = self.reader.next()
        if data:
            try:
                json_data = json.loads(data.strip('\n\r,'))
            except ValueError:
                raise StopIteration()
            record = {'id': json_data['id']}
            record['value'] = (dict((k, v) for (k, v) in
                               json_data['value'].iteritems()
                               if not k.startswith('_')))
            return record
        else:
            raise StopIteration()


class JSONWriter(migrator.Writer):
    def __init__(self, destination):
        if destination[0:2] == "//":
            destination = destination[2:]
        self.file = open(destination, 'w')

    def write(self, record):
        self.file.write(json.dumps(record) + '\n')



#
# Copyright 2012, Couchbase, Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

sources = [{'type':'couchbase', 'class':'CouchbaseReader',
            'example':'couchbase://bucket:password@example.com:8091/bucket'}]
destinations = [{'type':'couchbase', 'class':'CouchbaseWriter',\
    'example': 'couchbase://bucket:password@example.com:8091/bucket'}]

from urlparse import urlparse
import urllib
import couchbase
import migrator


class CouchbaseReader(migrator.Reader):
    def __init__(self, source):
        # couchbase://username:password@example.com:8091/bucket
        url = urlparse(source)
        self.username = url.username
        self.password = url.password
        self.host = url.hostname
        self.port = url.port
        self.bucket_name = url.path[1:]

        self.page_limit = 100

        cb = couchbase.Server(self.host + ":" + str(self.port),
                              username=self.username,
                              password=self.password)
        self.bucket = cb[self.bucket_name]

        self.items = self.bucket.view('_all_docs', limit=self.page_limit + 1,
                                      stale=False, reduce=False,
                                      include_docs=True)

    def __iter__(self):
        return self

    def next(self):
        if len(self.items) < 1:
            raise StopIteration()
        elif len(self.items) == 1:
            next_startkey = (urllib.quote_plus(self.items[0]['key']
                             .replace('"', '\\"').encode('utf-8')))
            next_startkey_docid = urllib.quote_plus(self.items[0]['doc']['_id']
                                                    .replace('"', '\\"')
                                                    .encode('utf-8'))
            self.items = self.bucket.view('_all_docs',
                                          limit=self.page_limit + 1,
                                          startkey=next_startkey,
                                          startkey_docid=next_startkey_docid,
                                          stale=False, reduce=False,
                                          include_docs=True)
            data = self.items.pop(0)
        else:
            data = self.items.pop(0)

        record = {'id': data['doc']['_id']}
        record['value'] = (dict((k, v) for (k, v) in data['doc'].iteritems()
                           if not k.startswith('$')))
        return record


class CouchbaseWriter(migrator.Writer):
    def __init__(self, destination):
        # couchbase://username:password@example.com:8091/bucket
        url = urlparse(destination)
        self.username = url.username
        self.password = url.password
        self.host = url.hostname
        self.port = url.port
        self.bucket_name = url.path[1:]

        self.verbose = False

        cb = couchbase.Server(self.host + ":" + str(self.port),
                              username=self.username,
                              password=self.password)
        self.bucket = cb[self.bucket_name]

    def write(self, record):
        record_save = record['value']
        record_save['_id'] = record['id'].encode('utf-8')
        self.bucket.save(record_save)



#
# Copyright 2012, Couchbase, Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

sources = [{'type':'csv', 'class':'CSVReader', 'example':'csv://<filename>'}]
destinations = []

import csv
import migrator


class CSVReader(migrator.Reader):
    def __init__(self, source):
        if source[0:2] == "//":
            source = source[2:]
        self.reader = csv.DictReader(open(source, 'rb'))
        self.reader.fieldnames[0] = 'id'

    def __iter__(self):
        return self

    def next(self):
        data = self.reader.next()
        if data:
            record = {'id': data['id']}
            record['value'] = (dict((k, v) for (k, v) in data.iteritems()
                               if k != 'id' and not k.startswith('_')))
            return record
        else:
            raise StopIteration()


class CSVWriter(migrator.Writer):
    pass



#!/usr/bin/env python

from copy import deepcopy

import couchdb
import httplib
import common
try: import simplejson as json
except ImportError: import json
import urllib
import unittest

import operator
import random
import time

HOST = "localhost:5984"
SET_NAME_LOCAL = "test_suite_set_view_view_merge_local"
SET_NAME_REMOTE = "test_suite_set_view_view_merge_remote"
NUM_PARTS = 4
NUM_DOCS = 5000

DDOC = {
    "_id": "_design/test",
    "language": "javascript",
    "views": {
        "mapview": {
            "map": "function(doc) { emit(doc.integer, doc.string); }"
        },
        "redview": {
            "map":  "function(doc) {" \
                "emit([doc.integer, doc.string], doc.integer);" \
                "emit([doc.integer + 1, doc.string], doc.integer + 1);" \
                "}",
            "reduce": "_sum"
        },
        "redview2": {
            "map":  "function(doc) {" \
                "emit(doc.integer, doc.string);" \
                "}",
            "reduce": "_count"
        }
    }
}


class TestViewMerge(unittest.TestCase):

    def setUp(self):
        server = couchdb.Server(url = "http://" + HOST)
        self._params_local = {
            "host": HOST,
            "ddoc": deepcopy(DDOC),
            "nparts": NUM_PARTS,
            "ndocs": NUM_DOCS,
            "setname": SET_NAME_LOCAL,
            "server": server
            }
        self._params_remote = {
            "host": HOST,
            "ddoc": deepcopy(DDOC),
            "nparts": NUM_PARTS,
            "start_id": NUM_DOCS,
            "ndocs": NUM_DOCS,
            "setname": SET_NAME_REMOTE,
            "server": server
            }
        # print "Creating databases"
        common.create_dbs(self._params_local)
        common.create_dbs(self._params_remote)
        common.populate(self._params_local)
        common.populate(self._params_remote)

        common.define_set_view(self._params_local, range(NUM_PARTS), [])
        common.define_set_view(self._params_remote, range(NUM_PARTS), [])
        # print "Databases created"


    def tearDown(self):
        # print "Deleting test data"
        common.create_dbs(self._params_local, True)
        common.create_dbs(self._params_remote, True)


    def test_view_merge(self):
        self.do_test_mapview(self._params_local, self._params_remote)
        self.do_test_redview(self._params_local, self._params_remote)
        self.do_test_redview2(self._params_local, self._params_remote)
        self.do_test_include_docs(self._params_local, self._params_remote)
        self.do_test_limit0(self._params_local, self._params_remote)
        self.do_test_limit(self._params_local, self._params_remote)
        self.do_test_skip(self._params_local, self._params_remote)
        self.do_test_filter(self._params_local, self._params_remote)
        self.do_test_start_end_key(self._params_local, self._params_remote)
        self.do_test_descending(self._params_local, self._params_remote)
        self.do_test_key(self._params_local, self._params_remote)
        self.do_test_keys(self._params_local, self._params_remote)
        self.do_test_set_view_outdated_local(self._params_local, self._params_remote)
        self.do_test_query_args(self._params_local, self._params_remote)
        self.do_test_debug_info(self._params_local, self._params_remote)
        self.do_test_replica_support()


    def set_spec(self, name, view, partitions):
        return (name, {"view" : "test/%s" % view,
                       "partitions" : partitions})


    def views_spec(self, specs, set_specs):
        specs_dict = dict(specs)
        specs_dict["sets"] = dict(set_specs)

        return {"views" : specs_dict}


    def merge_spec(self, server, specs, set_specs):
        return ("http://%s/_view_merge/" % server, self.views_spec(specs, set_specs))


    def do_query(self, server, spec, params = {}):
        params = deepcopy(params)
        params.setdefault("stale", "false")

        url = "http://%s/_view_merge/" % server
        if params:
            url += "?%s" % urllib.urlencode(params)

        conn = httplib.HTTPConnection(server)
        conn.request("POST", url,
                     headers = {"Content-Type": "application/json"},
                     body = json.dumps(spec))

        resp = conn.getresponse()
        body = json.loads(resp.read())
        conn.close()
        return (resp, body)


    def query(self, server, spec, params = {}):
        resp, body = self.do_query(server, spec, params)
        self.assertEqual(resp.status, 200, "View query response has status %d" % resp.status)
        return (resp, body)


    def do_test_mapview(self, local, remote):
        local_spec = self.set_spec(local["setname"], "mapview", range(local["nparts"]))
        remote_spec = self.set_spec(remote["setname"], "mapview", range(remote["nparts"]))
        remote_merge = self.merge_spec(remote["host"], [], [remote_spec])

        full_spec = self.views_spec([remote_merge], [local_spec])
        _, result = self.query(local["host"], full_spec)

        self.assertEqual(result["total_rows"], local["ndocs"] + remote["ndocs"],
                         "Total rows differs from %d" % (local["ndocs"] + remote["ndocs"]))
        self.assertEqual(len(result["rows"]), local["ndocs"] + remote["ndocs"],
                         "len(rows) from %d" % (local["ndocs"] + remote["ndocs"]))
        common.test_keys_sorted(result)


    def do_test_redview(self, local, remote):
        local_spec = self.set_spec(local["setname"], "redview", range(local["nparts"]))
        remote_spec = self.set_spec(remote["setname"], "redview", range(remote["nparts"]))
        remote_merge = self.merge_spec(remote["host"], [], [remote_spec])

        full_spec = self.views_spec([remote_merge], [local_spec])

        _, result = self.query(local["host"], full_spec)
        self.assertEqual(len(result["rows"]), 1,
            "Query returned invalid number of rows (a)")
        self.assertEqual(result["rows"][0]["value"], 100020000,
                         "Non-grouped reduce value is not 100020000")

        _, result = self.query(local["host"], full_spec, params={"group": True})
        self.assertEqual(len(result["rows"]), 20000,
                         "Query returned invalid number of rows (b)")

        _, result = self.query(local["host"], full_spec, params={"group_level": 1})
        self.assertEqual(len(result["rows"]), 10001,
                         "Query returned invalid number of rows (c)")


    def do_test_redview2(self, local, remote):
        local_spec = self.set_spec(local["setname"], "redview2", range(local["nparts"]))
        remote_spec = self.set_spec(remote["setname"], "redview2", range(remote["nparts"]))
        remote_merge = self.merge_spec(remote["host"], [], [remote_spec])

        full_spec = self.views_spec([remote_merge], [local_spec])

        expected_row_count = common.set_doc_count(local, range(local["nparts"])) + \
            common.set_doc_count(remote, range(remote["nparts"]))
        _, result = self.query(local["host"], full_spec, {"reduce": "false"})
        self.assertEqual(len(result["rows"]), expected_row_count,
            "redview2?reduce=false query returned %d rows" % expected_row_count)

        common.test_keys_sorted(result)

        _, result = self.query(local["host"], full_spec)
        self.assertEqual(len(result["rows"]), 1, "redview2?reduce=true query returned 1 row")
        self.assertEqual(result["rows"][0]["value"], expected_row_count,
                         "Non-grouped reduce value is %d" % expected_row_count)

        _, result = self.query(local["host"], full_spec, {"group": "true"})
        self.assertEqual(len(result["rows"]), expected_row_count,
            "redview2?group=true query returned %d rows" % expected_row_count)

        for i in xrange(0, len(result["rows"]) - 1):
            row1 = result["rows"][i]
            row2 = result["rows"][i + 1]
            self.assertTrue(row1["key"] < row2["key"], "row1.key < row2.key")
            self.assertEqual(row1["value"], 1, "row1.value == 1")
            self.assertEqual(row2["value"], 1, "row2.value == 1")


    def do_test_include_docs(self, local, remote):
        local_spec = self.set_spec(local["setname"], "mapview", range(local["nparts"]))
        remote_spec = self.set_spec(remote["setname"], "mapview", range(remote["nparts"]))
        remote_merge = self.merge_spec(remote["host"], [], [remote_spec])

        full_spec = self.views_spec([remote_merge], [local_spec])
        _, result = self.query(local["host"], full_spec, params={"include_docs" : True})

        self.assertEqual(result["total_rows"], local["ndocs"] + remote["ndocs"],
                         "Total rows differs from %d" % (local["ndocs"] + remote["ndocs"]))
        self.assertEqual(len(result["rows"]), local["ndocs"] + remote["ndocs"],
                         "length(rows) differs from %d" % (local["ndocs"] + remote["ndocs"]))
        self.assertFalse("errors" in result, "query result has no errors")
        common.test_keys_sorted(result)

        for row in result["rows"]:
            key = row["key"]
            doc = row["doc"]
            self.assertEqual(doc["integer"], key, "Invalid `doc` returned for %d key" % key)
            self.assertEqual(doc["string"], str(key), "Invalid `doc` returned for %d key" % key)

    def do_test_limit0(self, local, remote):
        local_spec = self.set_spec(local["setname"], "mapview", range(local["nparts"]))
        remote_spec = self.set_spec(remote["setname"], "mapview", range(remote["nparts"]))
        remote_merge = self.merge_spec(remote["host"], [], [remote_spec])

        full_spec = self.views_spec([remote_merge], [local_spec])
        _, result = self.query(local["host"], full_spec, params = {"limit" : 0})

        self.assertEqual(result["total_rows"], local["ndocs"] + remote["ndocs"],
                         "Total rows differs from %d" % (local["ndocs"] + remote["ndocs"]))

        self.assertEqual(len(result["rows"]), 0, "Invalid row count")

    def do_test_limit(self, local, remote):
        local_spec = self.set_spec(local["setname"], "mapview", range(local["nparts"]))
        remote_spec = self.set_spec(remote["setname"], "mapview", range(remote["nparts"]))
        remote_merge = self.merge_spec(remote["host"], [], [remote_spec])

        full_spec = self.views_spec([remote_merge], [local_spec])
        _, result = self.query(local["host"], full_spec, params = {"limit" : 10})

        self.assertEqual(result["total_rows"], local["ndocs"] + remote["ndocs"],
                         "Total rows differs from %d" % (local["ndocs"] + remote["ndocs"]))
        common.test_keys_sorted(result)

        self.assertEqual(len(result["rows"]), 10, "Invalid row count")

        for (i, row) in enumerate(result["rows"], 1):
            key = row["key"]
            self.assertEqual(key, i, "Got invalid key")


    def do_test_skip(self, local, remote):
        local_spec = self.set_spec(local["setname"], "mapview", range(local["nparts"]))
        remote_spec = self.set_spec(remote["setname"], "mapview", range(remote["nparts"]))
        remote_merge = self.merge_spec(remote["host"], [], [remote_spec])

        full_spec = self.views_spec([remote_merge], [local_spec])
        _, result = self.query(local["host"], full_spec,
                               params = {"limit" : 10, "skip" : 10})

        self.assertEqual(result["total_rows"], local["ndocs"] + remote["ndocs"],
                         "Total rows differs from %d" % (local["ndocs"] + remote["ndocs"]))
        common.test_keys_sorted(result)

        self.assertEqual(len(result["rows"]), 10, "Invalid row count")

        for (i, row) in enumerate(result["rows"], 11):
            key = row["key"]
            self.assertEqual(key, i, "Got invalid key")


    def do_test_filter(self, local, remote):
        local_spec = self.set_spec(local["setname"], "mapview", [0])
        remote_spec = self.set_spec(remote["setname"], "mapview", [remote["nparts"] - 1])
        remote_merge = self.merge_spec(remote["host"], [], [remote_spec])

        full_spec = self.views_spec([remote_merge], [local_spec])
        _, result = self.query(local["host"], full_spec)

        common.test_keys_sorted(result)

        for row in result["rows"]:
            key = row["key"]
            if key <= local["ndocs"]:
                self.assertEqual(key % local["nparts"], 1, "Got a document from filtered partition")
            else:
                self.assertEqual(key % remote["nparts"], 0, "Got a document from filtered partition")


    def do_test_start_end_key(self, local, remote):
        local_spec = self.set_spec(local["setname"], "mapview", range(local["nparts"]))
        remote_spec = self.set_spec(remote["setname"], "mapview", range(remote["nparts"]))
        remote_merge = self.merge_spec(remote["host"], [], [remote_spec])

        start_key = local["ndocs"] - 10
        end_key = local["ndocs"] + 10

        full_spec = self.views_spec([remote_merge], [local_spec])
        _, result = self.query(
            local["host"],
            full_spec,
            params = {"start_key" : start_key, "end_key" : end_key}
            )

        self.assertEqual(result["total_rows"], local["ndocs"] + remote["ndocs"],
                         "Total rows differs from %d" % (local["ndocs"] + remote["ndocs"]))
        common.test_keys_sorted(result)

        for row in result["rows"]:
            key = row["key"]
            self.assertTrue(key >= start_key and key <= end_key, "Got an item with invalid key %d" % key)


    def do_test_descending(self, local, remote):
        local_spec = self.set_spec(local["setname"], "mapview", range(local["nparts"]))
        remote_spec = self.set_spec(remote["setname"], "mapview", range(remote["nparts"]))
        remote_merge = self.merge_spec(remote["host"], [], [remote_spec])

        full_spec = self.views_spec([remote_merge], [local_spec])
        _, result = self.query(local["host"], full_spec, params = {"descending" : True})

        self.assertEqual(result["total_rows"], local["ndocs"] + remote["ndocs"],
                         "Total rows differs from %d" % (local["ndocs"] + remote["ndocs"]))
        common.test_keys_sorted(result, operator.gt)


    def do_test_key(self, local, remote):
        local_spec = self.set_spec(local["setname"], "mapview", range(local["nparts"]))
        remote_spec = self.set_spec(remote["setname"], "mapview", range(remote["nparts"]))
        remote_merge = self.merge_spec(remote["host"], [], [remote_spec])

        full_spec = self.views_spec([remote_merge], [local_spec])

        n = local["ndocs"] + remote["ndocs"]
        for i in xrange(100):
            wanted_key = random.randint(1, n)
            _, result = self.query(local["host"], full_spec, params = {"key" : wanted_key})

            self.assertEqual(result["total_rows"], local["ndocs"] + remote["ndocs"],
                             "Total rows differs from %d" % (local["ndocs"] + remote["ndocs"]))
            self.assertEqual(len(result["rows"]), 1, "Result has more than one row")

            key = result["rows"][0]["key"]
            self.assertEqual(key, wanted_key,
                             "Returned key %d differs from the requested %d" % (key, wanted_key))


    def do_test_keys(self, local, remote):
        local_spec = self.set_spec(local["setname"], "mapview", range(local["nparts"]))
        remote_spec = self.set_spec(remote["setname"], "mapview", range(remote["nparts"]))
        remote_merge = self.merge_spec(remote["host"], [], [remote_spec])

        n = local["ndocs"] + remote["ndocs"]
        wanted_keys = [random.randint(1, n) for i in xrange(100)]

        full_spec = self.views_spec([remote_merge], [local_spec])
        full_spec["keys"] = wanted_keys

        _, result = self.query(local["host"], full_spec, params = {"keys" : wanted_keys})

        self.assertEqual(result["total_rows"], local["ndocs"] + remote["ndocs"],
                         "Total rows differs from %d" % (local["ndocs"] + remote["ndocs"]))
        keys = set(row["key"] for row in result["rows"])

        self.assertEqual(set(wanted_keys), keys,
                         "Got response having excessive or missing some of the requested keys")


    def do_test_set_view_outdated_local(self, local, remote):
        local_spec = self.set_spec(local["setname"], "mapview", range(local["nparts"] + 1))
        remote_spec = self.set_spec(remote["setname"], "mapview", range(remote["nparts"]))
        remote_merge = self.merge_spec(remote["host"], [], [remote_spec])

        full_spec = self.views_spec([remote_merge], [local_spec])
        (resp, body) = self.do_query(local["host"], full_spec)
        self.assertTrue('error' in body and body.get('reason', 'set_view_outdated'))


    def do_test_set_view_outdated_remote(self, local, remote):
        local_spec = self.set_spec(local["setname"], "mapview", range(local["nparts"]))
        remote_spec = self.set_spec(remote["setname"], "mapview", range(remote["nparts"] + 1))
        remote_merge = self.merge_spec(remote["host"], [], [remote_spec])

        full_spec = self.views_spec([remote_merge], [local_spec])
        (resp, body) = self.do_query(local["host"], full_spec)
        self.assertTrue('error' in body and body.get('reason', 'set_view_outdated'))


    def do_test_query_args(self, local, remote):
        local_spec = self.set_spec(local["setname"], "mapview", range(local["nparts"]))
        remote_spec = self.set_spec(remote["setname"], "mapview", range(remote["nparts"]))
        remote_merge = self.merge_spec(remote["host"], [], [remote_spec])

        full_spec = self.views_spec([remote_merge], [local_spec])
        resp, result = self.do_query(local["host"], full_spec,
                                     {"connection_timeout": '"30000"'})

        self.assertEqual(resp.status, 400, "Return status is correct")
        self.assertEqual(result["error"], "bad_request", "Correct error")
        self.assertNotEqual(result["reason"].find("connection_timeout"), -1,
                            "Correct messsage")

        resp, result = self.do_query(local["host"], full_spec,
                                     {"on_error": '"foo"'})

        self.assertEqual(resp.status, 400, "Return status is correct")
        self.assertEqual(result["error"], "bad_request", "Correct error")
        self.assertNotEqual(result["reason"].find("on_error"), -1,
                            "Correct messsage")


    def do_test_debug_info(self, local, remote):
        local_spec = self.set_spec(local["setname"], "mapview", range(local["nparts"]))
        remote_spec = self.set_spec(remote["setname"], "mapview", range(remote["nparts"]))
        remote_merge = self.merge_spec(remote["host"], [], [remote_spec])

        full_spec = self.views_spec([remote_merge], [local_spec])
        _, result = self.query(local["host"], full_spec, {"debug": "true"})

        self.assertEqual(result["total_rows"], local["ndocs"] + remote["ndocs"],
                         "Total rows differs from %d" % (local["ndocs"] + remote["ndocs"]))
        self.assertEqual(len(result["rows"]), local["ndocs"] + remote["ndocs"],
                         "len(rows) from %d" % (local["ndocs"] + remote["ndocs"]))

        self.assertTrue("debug_info" in result, "Got debug_info field in response")
        debug_info = result["debug_info"]
        self.assertTrue(type(debug_info) == dict, "debug_info field is an object")
        self.assertEqual(len(debug_info), 2, "debug_info field has 2 fields")

        for (key, info) in debug_info.iteritems():
            self.assertTrue(type(info) == dict, "debug_info field is an object")

        common.test_keys_sorted(result)

        next_part = 0
        for row in result["rows"]:
            part = row["partition"]
            self.assertEqual(part, next_part, "row for key %d has right partition id field" % row["key"])
            next_part = (next_part + 1) % NUM_PARTS

        local_spec = self.set_spec(local["setname"], "redview", range(local["nparts"]))
        remote_spec = self.set_spec(remote["setname"], "redview", range(remote["nparts"]))
        remote_merge = self.merge_spec(remote["host"], [], [remote_spec])

        full_spec = self.views_spec([remote_merge], [local_spec])

        _, result = self.query(local["host"], full_spec, {"debug": "true"})
        self.assertEqual(len(result["rows"]), 1,
            "Query returned invalid number of rows (a)")
        self.assertEqual(result["rows"][0]["value"], 100020000,
                         "Non-grouped reduce value is not 100020000")
        self.assertTrue("debug_info" in result, "Got debug_info field in response")
        debug_info = result["debug_info"]
        self.assertTrue(type(debug_info) == dict, "debug_info field is an object")
        self.assertEqual(len(debug_info), 2, "debug_info field has 2 fields")

        for (key, info) in debug_info.iteritems():
            self.assertTrue(type(info) == dict, "debug_info field is an object")

        self.assertFalse("partition" in result["rows"][0], "reduce row has no partition field")


    def do_test_replica_support(self):
        server = couchdb.Server(url = "http://" + HOST)
        local_params = {
            "host": HOST,
            "ddoc": deepcopy(DDOC),
            "nparts": NUM_PARTS,
            "ndocs": 50000,
            "setname": SET_NAME_LOCAL + "_2",
            "server": server
            }
        remote_params = {
            "host": HOST,
            "ddoc": deepcopy(DDOC),
            "nparts": NUM_PARTS,
            "start_id": 50000,
            "ndocs": 50000,
            "setname": SET_NAME_REMOTE + "_2",
            "server": server
            }
        # print "Creating databases"
        common.create_dbs(local_params)
        common.create_dbs(remote_params)
        common.populate(local_params)
        common.populate(remote_params)
        # print "Databases created"

        common.define_set_view(local_params, [0, 1], [], True)
        common.define_set_view(remote_params, [2, 3], [], True)
        common.add_replica_partitions(local_params, [2, 3])
        common.add_replica_partitions(remote_params, [0, 1])

        local_spec = self.set_spec(local_params["setname"], "mapview", [0, 1])
        remote_spec = self.set_spec(remote_params["setname"], "mapview", [2, 3])
        remote_merge = self.merge_spec(remote_params["host"], [], [remote_spec])

        # print "Triggering view update"
        full_spec = self.views_spec([remote_merge], [local_spec])
        _, result = self.query(local_params["host"], full_spec, {"limit": "1", "stale": "false"})

        info = common.get_set_view_info(local_params)
        seconds = 0
        while info["replica_group_info"]["stats"]["full_updates"] < 1:
            time.sleep(3)
            seconds += 3
            if seconds > 900:
                raise(Exception("timeout waiting for replica group full update"))
            info = common.get_set_view_info(local_params)

        info = common.get_set_view_info(remote_params)
        seconds = 0
        while info["replica_group_info"]["stats"]["full_updates"] < 1:
            time.sleep(3)
            seconds += 3
            if seconds > 900:
                raise(Exception("timeout waiting for replica group full update"))
            info = common.get_set_view_info(remote_params)

        local_spec = self.set_spec(local_params["setname"], "mapview", [0, 1])
        remote_spec = self.set_spec(remote_params["setname"], "mapview", [2, 3])
        remote_merge = self.merge_spec(remote_params["host"], [], [remote_spec])
        full_spec = self.views_spec([remote_merge], [local_spec])
        _, result = self.query(local_params["host"], full_spec, {"stale": "false"})
        self.assertEqual(len(result["rows"]), 50000, "response has 50000 rows")

        # print "Activating replica partitions"
        common.set_partition_states(local_params, active = [2, 3])
        common.set_partition_states(remote_params, active = [0, 1])

        # print "Querying with ?debug=true"
        local_spec = self.set_spec(local_params["setname"], "mapview", [0, 1, 2, 3])
        remote_spec = self.set_spec(remote_params["setname"], "mapview", [0, 1, 2, 3])
        remote_merge = self.merge_spec(remote_params["host"], [], [remote_spec])
        full_spec = self.views_spec([remote_merge], [local_spec])
        _, result = self.query(local_params["host"], full_spec, {"stale": "ok", "debug": "true"})

        self.assertEqual(len(result["rows"]), 100000, "response has 100000 rows")
        self.assertTrue("debug_info" in result, "result has debug info")
        self.assertTrue("local" in result["debug_info"], "result have debug info for local node")
        remote_node = "http://%s/_view_merge/" % remote_params["host"]
        self.assertTrue(remote_node in result["debug_info"], "result have debug info for remote node")

        next_key = 1
        next_part = 0
        i = 0
        while next_key < 100000:
            row = result["rows"][i]
            self.assertEqual(row["key"], next_key, "correct row key at iteration %d" % (i + 1))
            self.assertEqual(row["value"], str(next_key), "correct row value at iteration %d" % (i + 1))
            self.assertEqual(row["partition"], next_part, "correct row partition at iteration %d" % (i + 1))
            node = "local"
            if next_key > 50000:
                node = "http://%s/_view_merge/" % remote_params["host"]
            self.assertEqual(row["node"], node, "correct row node at iteration %d" % (i + 1))
            i += 1
            next_key += 1
            next_part = (next_part + 1) % 4

        # print "Deleting databases"
        common.create_dbs(local_params, True)
        common.create_dbs(remote_params, True)




#!/usr/bin/env python

#
# Copyright 2012, Couchbase, Inc.
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
import getopt
import couchbase.migrator

# sources: json, csv, sqlite, couchdb, couchbase, membase, memcached,
# membase-sqlite
# dest: json (chunked), couchbase


def usage(err=None):
    if err:
        print "Error: %s\n" % err
        r = 1
    else:
        r = 0

    print """\
Syntax: couchbase-migrator [options]

Options:
 -h, --help

 -s <source>, --source=<source>
     Data source that will be imported from

 -d <destination>, --destination=<destination>
     Data destination that will be exported to
"""
    """
 -o, --overwrite
     Overwrite any items that are in the destination

 --dry-run
     Do not write any data, just show what would be written

 -v, --verbose

 -q, --quiet

"""
    print "Sources:"
    for source in couchbase.migrator.sources:
        print " " + source['type']
        print "     " + source['example']
        print

    print

    print "Destinations:"
    for destination in couchbase.migrator.destinations:
        print " " + destination['type']
        print "     " + destination['example']
        print

    sys.exit(r)


class Config(object):
    def __init__(self):
        self.source = ''
        self.destination = ''
        self.overwrite = False
        self.dryrun = False
        self.verbose = False
        self.quiet = False


def parse_args(argv):
    config = Config()
    try:
        opts, args = getopt.getopt(argv[1:],
                                     'hs:d:ovq', [
                'help',
                'source=',
                'destination=',
                'overwrite',
                'dry-run',
                'verbose',
                'quiet',
                ])
        for o, a in opts:
            if o == '-h' or o == '--help':
                usage()
            elif o == '-s' or o == '--source':
                config.source = a
            elif o == '-d' or o == '--destination':
                config.destination = a
            elif o == '-o' or o == '--overwrite':
                config.overwrite = True
            elif o == '--dry-run':
                config.dryrun = True
            elif o == '-v' or o == '--verbose':
                config.verbose = True
            elif o == '-q' or o == '--quiet':
                config.quiet = True

        msg = ""
        if not config.source or not config.destination:
            usage("missing source or destination")

    except IndexError:
        usage()
    except getopt.GetoptError, err:
        usage(err)

    return config


if __name__ == "__main__":
    config = parse_args(sys.argv)

    count = 0

    reader = couchbase.migrator.reader(config.source)
    writer = couchbase.migrator.writer(config.destination)
    for record in reader:
        writer.write(record)
        count += 1

    reader.close()
    writer.close()

    print 'migrated %s items' % (count)



#!/usr/bin/python

try: import simplejson as json
except ImportError: import json
import couchdb
import httplib
import urllib
import time
import common
import unittest


HOST = "localhost:5984"
SET_NAME = "test_suite_set_view"
NUM_PARTS = 16
NUM_DOCS = 800000
DDOC = {
    "_id": "_design/test",
    "language": "javascript",
    "views": {
        "mapview": {
            "map": "function(doc) { emit(doc.integer, doc.string); }"
        }
    }
}


# Issue a burst of partition state changes, and verify query results and
# index state are correct.
class TestBurstStateUpdates(unittest.TestCase):

    def setUp(self):
        self._params = {
            "host": HOST,
            "ddoc": DDOC,
            "nparts": NUM_PARTS,
            "ndocs": NUM_DOCS,
            "setname": SET_NAME,
            "server": couchdb.Server(url = "http://" + HOST)
            }
        # print "Creating %d databases, %d documents per database" % (NUM_PARTS, NUM_DOCS / NUM_PARTS)
        common.create_dbs(self._params)
        common.populate(self._params)
        # print "Databases created"
        # print "Configuring set view with:"
        # print "\tmaximum of 16 partitions"
        # print "\tactive partitions = []"
        # print "\tpassive partitions = []"
        common.define_set_view(self._params, [], [])


    def tearDown(self):
        # print "Deleting test data"
        common.create_dbs(self._params, True)


    def test_burst_state_updates(self):
        state_transitions = [
            {"active": [], "passive": [], "cleanup": []},
            {"active": [0], "passive": [], "cleanup": []},
            {"active": [1], "passive": [], "cleanup": []},
            {"active": [2], "passive": [], "cleanup": []},
            {"active": [3], "passive": [], "cleanup": []},
            {"active": [4], "passive": [], "cleanup": []},
            {"active": [5], "passive": [], "cleanup": []},
            {"active": [6], "passive": [], "cleanup": []},
            {"active": [7], "passive": [], "cleanup": []},
            {"active": [8], "passive": [], "cleanup": []},
            {"active": [9], "passive": [], "cleanup": []},
            {"active": [], "passive": [], "cleanup": []},
            {"active": [10], "passive": [], "cleanup": []},
            {"active": [11], "passive": [], "cleanup": []},
            {"active": [12], "passive": [], "cleanup": []},
            {"active": [13], "passive": [], "cleanup": []},
            {"active": [14], "passive": [], "cleanup": []},
            {"active": [15], "passive": [], "cleanup": []},
            {"active": [], "passive": [], "cleanup": [0]},
            {"active": [], "passive": [], "cleanup": [1]},
            {"active": [], "passive": [], "cleanup": [2]},
            {"active": [], "passive": [], "cleanup": [3]},
            {"active": [], "passive": [], "cleanup": [4]},
            {"active": [], "passive": [], "cleanup": [5]},
            {"active": [], "passive": [], "cleanup": [6]},
            {"active": [], "passive": [], "cleanup": [7]},
            {"active": [], "passive": [0], "cleanup": []},
            {"active": [], "passive": [1], "cleanup": []},
            {"active": [], "passive": [2], "cleanup": []},
            {"active": [], "passive": [3], "cleanup": []},
            {"active": [], "passive": [4], "cleanup": []},
            {"active": [], "passive": [5], "cleanup": []},
            {"active": [], "passive": [6], "cleanup": []},
            {"active": [], "passive": [7], "cleanup": []},
            {"active": [], "passive": [], "cleanup": [8]},
            {"active": [], "passive": [], "cleanup": []},
            {"active": [], "passive": [], "cleanup": [9]},
            {"active": [], "passive": [], "cleanup": [10]},
            {"active": [], "passive": [], "cleanup": [11]},
            {"active": [], "passive": [], "cleanup": [12]},
            {"active": [], "passive": [], "cleanup": [13]},
            {"active": [], "passive": [], "cleanup": [14]},
            {"active": [], "passive": [], "cleanup": [15]},
            {"active": [], "passive": [], "cleanup": []},
            {"active": [0], "passive": [], "cleanup": []},
            {"active": [1], "passive": [], "cleanup": []},
            {"active": [2], "passive": [], "cleanup": []},
            {"active": [3], "passive": [], "cleanup": []},
            {"active": [4], "passive": [], "cleanup": []},
            {"active": [], "passive": [], "cleanup": []},
            {"active": [5], "passive": [], "cleanup": []},
            {"active": [6], "passive": [], "cleanup": []},
            {"active": [7], "passive": [], "cleanup": []},
            {"active": [], "passive": [8], "cleanup": []},
            {"active": [], "passive": [9], "cleanup": []},
            {"active": [], "passive": [10], "cleanup": []},
            {"active": [], "passive": [11], "cleanup": []},
            {"active": [], "passive": [], "cleanup": []},
            {"active": [], "passive": [12], "cleanup": []},
            {"active": [], "passive": [], "cleanup": []},
            {"active": [], "passive": [13], "cleanup": []},
            {"active": [], "passive": [14], "cleanup": []},
            {"active": [], "passive": [15], "cleanup": []},
            {"active": [8], "passive": [], "cleanup": []},
            {"active": [9], "passive": [], "cleanup": []},
            {"active": [10], "passive": [], "cleanup": []},
            {"active": [11], "passive": [], "cleanup": []},
            {"active": [12], "passive": [], "cleanup": []},
            {"active": [13], "passive": [], "cleanup": []},
            {"active": [], "passive": [], "cleanup": []},
            {"active": [14], "passive": [], "cleanup": []},
            {"active": [15], "passive": [], "cleanup": []},
            {"active": [], "passive": [], "cleanup": []},
            ]

        self.do_test_without_updates(state_transitions)
        self.do_test_with_updates(state_transitions)



    # trigger the state changes without updates in between
    def do_test_without_updates(self, state_transitions):

        # print "Triggering state changes without queries in between"
        for state in state_transitions:
            common.set_partition_states(
                params = self._params,
                active = state["active"],
                passive = state["passive"],
                cleanup = state["cleanup"]
                )

        group_stats = self.verify_final_index_state(state_transitions)
        self.assertEqual(group_stats["stopped_updates"], 0, "0 updater interruptions")
        self.assertEqual(group_stats["full_updates"], 1, "Got 1 full update")



    # trigger the state changes with updates in between
    # (queries with ?stale=update_after)
    def do_test_with_updates(self, state_transitions):

        # start from a clean state always
        common.set_partition_states(
            params = self._params,
            cleanup = range(self._params["nparts"])
            )
        # print "Cleaning set view via compaction"
        common.compact_set_view(self._params)

        (resp, view_result) = common.query(self._params, "mapview")
        self.assertEqual(view_result["total_rows"], 0, "total_rows is 0 after compaction")
        self.assertEqual(len(view_result["rows"]), 0, "0 rows returned after compaction")
        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        self.assertEqual(info["update_seqs"], {}, "right update seqs list")

        # print "Triggering state changes with queries (?stale=updater_after) in between"
        for state in state_transitions:
            common.set_partition_states(
                params = self._params,
                active = state["active"],
                passive = state["passive"],
                cleanup = state["cleanup"]
                )
            time.sleep(0.5)
            common.query(self._params, "mapview", {"stale": "update_after", "limit": "100"})

        group_stats = self.verify_final_index_state(state_transitions)
        self.assertTrue(group_stats["stopped_updates"] > 0, "Got at least 1 updater interruption")
        self.assertEqual(group_stats["full_updates"], 2, "Got 2 full updates")
        self.assertEqual(group_stats["compactions"], 1, "Got 1 compaction")
        self.assertTrue(group_stats["cleanups"] >= 1, "Got at least 1 full cleanup")



    def verify_final_index_state(self, state_transitions):
        # print "Verifying final view state"
        doc_count = common.set_doc_count(self._params)
        (resp, view_result) = common.query(self._params, "mapview")

        # print "Final view result (%d rows):  %s" % \
        #    (len(view_result["rows"]), json.dumps(view_result, sort_keys = True, indent = 4))

        self.assertEqual(view_result["total_rows"], doc_count,
                         "Query returned %d total_rows" % doc_count)
        self.assertEqual(len(view_result["rows"]), doc_count,
                         "Query returned %d rows" % doc_count)

        self.assertEqual(view_result["rows"][0]["key"], 1, "First key is 1")
        self.assertEqual(view_result["rows"][-1]["key"], doc_count, "Last key is %d" % doc_count)
        common.test_keys_sorted(view_result)

        # print "Verifying group info"
        info = common.get_set_view_info(self._params)
        # print "Final group info:  %s" % json.dumps(info, sort_keys = True, indent = 4)
        self.assertEqual(
            info["active_partitions"],
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15],
            "right active partitions list"
            )
        self.assertEqual(info["passive_partitions"], [], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")

        for i in [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]:
            expected_seq = common.partition_update_seq(self._params, i)
            self.assertEqual(info["update_seqs"][str(i)], expected_seq,
                             "right update seq number (%d) for partition %d" % (expected_seq, i + 1))

        return info["stats"]



__all__ = [
    'reader', 'writer',
]

from migrator_csv import CSVReader, CSVWriter
from migrator_json import JSONReader, JSONWriter
from migrator_couchdb import CouchdbReader, CouchdbWriter
from migrator_couchbase import CouchbaseReader, CouchbaseWriter
from migrator_dir import DirReader, DirWriter
from migrator_zip import ZipReader, ZipWriter

sources = []
destinations = []

sources.extend(migrator_couchbase.sources)
sources.extend(migrator_csv.sources)
sources.extend(migrator_json.sources)
sources.extend(migrator_couchdb.sources)
sources.extend(migrator_dir.sources)
sources.extend(migrator_zip.sources)

destinations.extend(migrator_couchbase.destinations)
destinations.extend(migrator_csv.destinations)
destinations.extend(migrator_json.destinations)
destinations.extend(migrator_couchdb.destinations)
destinations.extend(migrator_dir.destinations)
destinations.extend(migrator_zip.destinations)


def reader(loc):
    kind, fp = loc.split(':', 1)
    if kind.lower() == 'csv':
        return CSVReader(fp)
    elif kind.lower() == 'json':
        return JSONReader(fp)
    elif kind.lower() == 'couchdb':
        return CouchdbReader(fp)
    elif kind.lower() == 'couchbase':
        return CouchbaseReader(fp)
    elif kind.lower() == 'dir':
        return DirReader(fp)
    elif kind.lower() == 'zip':
        return ZipReader(fp)


def writer(loc):
    kind, fp = loc.split(':', 1)
    if kind.lower() == 'csv':
        return CSVWriter(fp)
    elif kind.lower() == 'json':
        return JSONWriter(fp)
    elif kind.lower() == 'couchdb':
        return CouchdbWriter(fp)
    elif kind.lower() == 'couchbase':
        return CouchbaseWriter(fp)
    elif kind.lower() == 'dir':
        return DirWriter(fp)
    elif kind.lower() == 'zip':
        return ZipWriter(fp)



#!/usr/bin/python

try: import simplejson as json
except ImportError: import json
import couchdb
import httplib
import urllib
import time
import common
import unittest

HOST = "localhost:5984"
SET_NAME = "test_suite_set_view"
NUM_PARTS = 4
NUM_DOCS = 200000
DDOC = {
    "_id": "_design/test",
    "language": "javascript",
    "views": {
        "mapview": {
            "map": "function(doc) { emit(doc.integer, doc.string); }"
        }
    }
}


class TestPassivePartitionsUpdate(unittest.TestCase):

    def setUp(self):
        self._params = {
            "host": HOST,
            "ddoc": DDOC,
            "nparts": NUM_PARTS,
            "ndocs": NUM_DOCS,
            "setname": SET_NAME,
            "server": couchdb.Server(url = "http://" + HOST)
            }


    def tearDown(self):
        # print "Deleting test data"
        common.create_dbs(self._params, True)


    def test_updates(self):
        self.do_test_updates()
        self.do_test_set_passive_partitions_when_updater_is_running()


    # Verify that we have both active and passive partitions, the updater
    # first indexes the active partitions and then after it indexes the
    # documents from the passive partitions. Clients querying with ?stale=false
    # are unblocked as soon as the indexing of the active partitions is done.
    # This test verifies that property.
    def do_test_updates(self):
        # print "Creating databases"
        common.create_dbs(self._params)
        common.populate(self._params)
        # print "Databases created"
        # print "Configuring set view with 1 active partition and 3 passive partitions"
        common.define_set_view(self._params, [0], [1, 2, 3])

        # print "Querying map view"
        (resp, view_result) = common.query(self._params, "mapview", {"limit": "20"})

        # print "Grabbing group info"
        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [1, 2, 3], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        self.assertEqual(info["update_seqs"]["0"], (self._params["ndocs"] / 4), "All data from partition 1 was indexed")

        sum_passive_seqs = 0
        for i in [1, 2, 3]:
            sum_passive_seqs += info["update_seqs"][str(i)]

        self.assertTrue(sum_passive_seqs < (self._params["ndocs"] - (self._params["ndocs"] / 4)),
                        "Passive partitions are still being indexed")
        self.assertEqual(info["updater_running"], True, "View updater still running")
        self.assertEqual(info["updater_state"], "updating_passive", "View updater in state 'updating_passive'")

        # print "Verifying view query response"
        self.assertEqual(len(view_result["rows"]), 20, "Query returned 20 rows")

        common.test_keys_sorted(view_result)
        self.assertEqual(view_result["rows"][0]["key"], 1, "First row has key 1")
        last_key_expected = range(1, self._params["ndocs"], self._params["nparts"])[19]
        self.assertEqual(view_result["rows"][-1]["key"], last_key_expected, "Last row has key %d" % last_key_expected)

        # print "Waiting for view updater to finish"
        while True:
            info = common.get_set_view_info(self._params)
            if not info["updater_running"]:
                break
            else:
                time.sleep(3)

        # print "Grabbing group info"
        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [1, 2, 3], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        for i in [0, 1, 2, 3]:
            self.assertEqual(info["update_seqs"][str(i)], (self._params["ndocs"] / 4),
                             "Right update seq for partition %d" % (i + 1))
        self.assertEqual(info["updater_running"], False, "View updater not running anymore")
        self.assertEqual(info["updater_state"], "not_running", "View updater not running anymore")

        # print "Querying view again and validating the response"
        (resp, view_result) = common.query(self._params, "mapview")

        self.assertEqual(len(view_result["rows"]), (self._params["ndocs"] / 4), "Query returned %d rows" % (self._params["ndocs"] / 4))

        common.test_keys_sorted(view_result)
        self.assertEqual(view_result["rows"][0]["key"], 1, "First row has key 1")
        last_key_expected = range(1, self._params["ndocs"], self._params["nparts"])[-1]
        self.assertEqual(view_result["rows"][-1]["key"], last_key_expected, "Last row has key %d" % last_key_expected)


    def do_test_set_passive_partitions_when_updater_is_running(self):
        # print "Re-creating databases"
        common.create_dbs(self._params)
        common.populate(self._params)
        # print "Configuring set view with all partitions active"
        common.define_set_view(self._params, [0, 1, 2, 3], [])

        # print "Querying map view in steady state with ?stale=update_after"
        (resp, view_result) = common.query(self._params, "mapview", {"stale": "update_after"})

        self.assertEqual(len(view_result["rows"]), 0, "Received empty row set")
        self.assertEqual(view_result["total_rows"], 0, "Received empty row set")

        # print "Marking partition 4 as passive"
        common.set_partition_states(self._params, passive = [3])

        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [3], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")

        # print "Waiting for the set view updater to finish"
        iterations = 0
        while True:
            info = common.get_set_view_info(self._params)
            if info["updater_running"]:
                iterations += 1
            else:
                break

        self.assertTrue(iterations > 0, "Updater was running when partition 4 was set to passive")
        # print "Verifying set view group info"
        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [3], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [],  "right cleanup partitions list")

        # print "Querying map view again"
        (resp, view_result) = common.query(self._params, "mapview")

        doc_count = common.set_doc_count(self._params)
        expected_row_count = common.set_doc_count(self._params, [0, 1, 2])
        self.assertEqual(view_result["total_rows"], doc_count, "Query returned %d total_rows" % doc_count)
        self.assertEqual(len(view_result["rows"]), expected_row_count, "Query returned %d rows" % expected_row_count)

        common.test_keys_sorted(view_result)

        all_keys = {}
        for r in view_result["rows"]:
            all_keys[r["key"]] = True

        for key in xrange(4, doc_count, self._params["nparts"]):
            self.assertFalse(key in all_keys,
                             "Key %d not in result after partition 4 was set to passive" % key)

        # print "Adding 2 new documents to partition 4"
        server = self._params["server"]
        db4 = server[self._params["setname"] + "/3"]
        new_doc1 = {"_id": "999999999", "integer": 999999999, "string": "999999999"}
        new_doc2 = {"_id": "000", "integer": -1111, "string": "000"}
        db4.save(new_doc1)
        db4.save(new_doc2)

        # print "Querying map view again"
        (resp, view_result2) = common.query(self._params, "mapview")
        self.assertEqual(view_result2["rows"], view_result["rows"], "Same result set as before")

        # print "Verifying set view group info"
        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [3], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [],  "right cleanup partitions list")

        total_doc_count = common.set_doc_count(self._params)
        # print "Changing partition 4 from passive to active"
        common.set_partition_states(self._params, active = [3])

        # print "Querying map view again"
        (resp, view_result) = common.query(self._params, "mapview")

        self.assertEqual(view_result["total_rows"], total_doc_count, "total_rows is %d" % total_doc_count)
        self.assertEqual(len(view_result["rows"]), total_doc_count, "number of rows returned is %d" % total_doc_count)
        common.test_keys_sorted(view_result)

        self.assertEqual(view_result["rows"][0]["key"], new_doc2["integer"], "new_doc2 reflected in result set at first row")
        self.assertEqual(view_result["rows"][-1]["key"], new_doc1["integer"], "new_doc1 reflected in result set at last row")

        # print "Verifying set view group info"
        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2, 3], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [],  "right cleanup partitions list")
        for i in [0, 1, 2, 3]:
            expected_seq = common.partition_update_seq(self._params, i)
            self.assertEqual(info["update_seqs"][str(i)], expected_seq,
                             "right update seq number (%d) for partition %d" % (expected_seq, i + 1))



#!/usr/bin/env python
"""
Example script for tap.py.

Copyright (c) 2010  Dustin Sallings <dustin@spy.net>
"""

import os
import sys
import string
import asyncore
import signal
import getopt

import mc_bin_server
import mc_bin_client

from memcacheConstants import REQ_MAGIC_BYTE, RES_MAGIC_BYTE
from memcacheConstants import REQ_PKT_FMT, RES_PKT_FMT, MIN_RECV_PACKET
from memcacheConstants import SET_PKT_FMT, DEL_PKT_FMT, INCRDECR_RES_FMT

import memcacheConstants

import tap

def usage(err=0):
    print >> sys.stderr, """
Usage: %s [-u bucket_user [-p bucket_password]] host:port [... hostN:portN]

Example:
  %s -u user_profiles -p secret9876 membase-01:11210 membase-02:11210
""" % (os.path.basename(sys.argv[0]),
       os.path.basename(sys.argv[0]))
    sys.exit(err)

def parse_args(args):
    user = None
    pswd = None

    try:
        opts, args = getopt.getopt(args, 'hu:p:', ['help'])
    except getopt.GetoptError, e:
        usage("ERROR: " + e.msg)

    for (o, a) in opts:
        if o == '--help' or o == '-h':
            usage()
        elif o == '-u':
            user = a
        elif o == '-p':
            pswd = a
        else:
            usage("ERROR: unknown option - " + o)

    if not args or len(args) < 1:
        usage("ERROR: missing at least one host:port to TAP")

    return user, pswd, args

def signal_handler(signal, frame):
    print 'Tap stream terminated by user'
    sys.exit(0)

def buildGoodSet(goodChars=string.printable, badChar='?'):
    """Build a translation table that turns all characters not in goodChars
    to badChar"""
    allChars=string.maketrans("", "")
    badchars=string.translate(allChars, allChars, goodChars)
    rv=string.maketrans(badchars, badChar * len(badchars))
    return rv

# Build a translation table that includes only characters
transt=buildGoodSet()

def abbrev(v, maxlen=30):
    if len(v) > maxlen:
        return v[:maxlen] + "..."
    else:
        return v

def keyprint(v):
    return string.translate(abbrev(v), transt)

def mainLoop(serverList, cb, opts={}, user=None, pswd=None):
    """Run the given callback for each tap message from any of the
    upstream servers.

    loops until all connections drop
    """
    signal.signal(signal.SIGINT, signal_handler)

    connections = (tap.TapDescriptor(a) for a in serverList)
    tap.TapClient(connections, cb, opts=opts, user=user, pswd=pswd)
    asyncore.loop()

if __name__ == '__main__':
    user, pswd, args = parse_args(sys.argv[1:])

    def cb(identifier, cmd, extra, key, vb, val, cas):
        print "%s: ``%s'' (vb:%d) -> ``%s'' (%d bytes from %s)" % (
            memcacheConstants.COMMAND_NAMES[cmd],
            key, vb, keyprint(val), len(val), identifier)

    # This is an example opts parameter to do future-only tap:
    opts = {memcacheConstants.TAP_FLAG_BACKFILL: 0xffffffff}
    # If you omit it, or supply a past time_t value for backfill, it
    # will get all data.
    opts = {}

    mainLoop(args, cb, opts, user, pswd)



#!/usr/bin/env python

import csv
import logging
import os
import simplejson as json
import sys

import memcacheConstants
import pump

def number_try_parse(str):
    for func in (int, float):
        try:
            return func(str)
        except ValueError:
            pass

    return str

class CSVSource(pump.Source):
    """Reads csv file, where first line is field names and one field
       should be 'id'."""

    def __init__(self, opts, spec, source_bucket, source_node,
                 source_map, sink_map, ctl, cur):
        super(CSVSource, self).__init__(opts, spec, source_bucket, source_node,
                                        source_map, sink_map, ctl, cur)
        self.done = False
        self.r = None # An iterator of csv.reader()

    @staticmethod
    def can_handle(opts, spec):
        return spec.endswith(".csv") and os.path.isfile(spec)

    @staticmethod
    def check(opts, spec):
        return 0, {'spec': spec,
                   'buckets': [{'name': os.path.basename(spec),
                                'nodes': [{'hostname': 'N/A'}]}]}

    @staticmethod
    def provide_design(opts, source_spec, source_bucket, source_map):
        return 0, None

    def provide_batch(self):
        if self.done:
            return 0, None

        if not self.r:
            try:
                self.r = csv.reader(open(self.spec, 'rU'))
                self.fields = self.r.next()
                if not 'id' in self.fields:
                    return ("error: no 'id' field in 1st line of csv: %s" %
                            (self.spec)), None
            except StopIteration:
                return ("error: could not read 1st line of csv: %s" %
                        (self.spec)), None
            except IOError, e:
                return ("error: could not open csv: %s; exception: %s" %
                        (self.spec, e)), None

        batch = pump.Batch(self)

        batch_max_size = self.opts.extra['batch_max_size']
        batch_max_bytes = self.opts.extra['batch_max_bytes']

        cmd = memcacheConstants.CMD_TAP_MUTATION
        vbucket_id = 0x0000ffff
        cas, exp, flg = 0, 0, 0

        while (self.r and
               batch.size() < batch_max_size and
               batch.bytes < batch_max_bytes):
            try:
                vals = self.r.next()
                doc = {}
                for i, field in enumerate(self.fields):
                    if field == 'id':
                        doc[field] = vals[i]
                    else:
                        doc[field] = number_try_parse(vals[i])
                doc_json = json.dumps(doc)
                msg = (cmd, vbucket_id, doc['id'], flg, exp, cas, '', doc_json)
                batch.append(msg, len(doc))
            except StopIteration:
                self.done = True
                self.r = None

        if batch.size() <= 0:
            return 0, None
        return 0, batch


class CSVSink(pump.Sink):
    """Emits batches to stdout in CSV format."""

    def __init__(self, opts, spec, source_bucket, source_node,
                 source_map, sink_map, ctl, cur):
        super(CSVSink, self).__init__(opts, spec, source_bucket, source_node,
                                      source_map, sink_map, ctl, cur)
        self.writer = None

    @staticmethod
    def can_handle(opts, spec):
        if spec == "csv:":
            opts.threads = 1 # Force 1 thread to not overlap stdout.
            return True
        return False

    @staticmethod
    def check(opts, spec, source_map):
        return 0, None

    @staticmethod
    def consume_design(opts, sink_spec, sink_map,
                       source_bucket, source_map, source_design):
        if source_design:
            logging.warn("warning: cannot save bucket design"
                         " on a CSV destination")
        return 0

    def consume_batch_async(self, batch):
        if not self.writer:
            self.writer = csv.writer(sys.stdout)
            self.writer.writerow(['id', 'flags', 'expiration', 'cas', 'value'])

        for msg in batch.msgs:
            cmd, vbucket_id, key, flg, exp, cas, meta, val = msg
            if self.skip(key, vbucket_id):
                continue

            try:
                if cmd == memcacheConstants.CMD_TAP_MUTATION:
                    self.writer.writerow([key, flg, exp, cas, val])
                elif cmd == memcacheConstants.CMD_TAP_DELETE:
                    pass
                elif cmd == memcacheConstants.CMD_GET:
                    pass
                else:
                    return "error: CSVSink - unknown cmd: " + str(cmd), None
            except IOError:
                return "error: could not write csv to stdout", None

        future = pump.SinkBatchFuture(self, batch)
        self.future_done(future, 0)
        return 0, future



#!/usr/bin/python

try: import simplejson as json
except ImportError: import json
import couchdb
import httplib
import urllib
import common
import time
import unittest


HOST = "localhost:5984"
SET_NAME = "test_suite_set_view_compact"
NUM_PARTS = 4
NUM_DOCS = 20000
DDOC = {
    "_id": "_design/test",
    "language": "javascript",
    "views": {
        "mapview1": {
            "map": "function(doc) { emit(doc.integer, doc.string); }"
        }
    }
}



class TestStale(unittest.TestCase):

    def setUp(self):
        self._params = {
            "host": HOST,
            "ddoc": DDOC,
            "nparts": NUM_PARTS,
            "ndocs": NUM_DOCS,
            "setname": SET_NAME,
            "server": couchdb.Server(url = "http://" + HOST)
            }
        # print "Creating %d databases, %d documents per database" % (NUM_PARTS, NUM_DOCS / NUM_PARTS)
        common.create_dbs(self._params)
        common.populate(self._params)
        # print "Databases created"
        # print "Configuring set view with:"
        # print "\tmaximum of 8 partitions"
        # print "\tactive partitions = [0, 1, 2, 3]"
        # print "\tpassive partitions = []"
        common.define_set_view(self._params, [0, 1, 2, 3], [])


    def tearDown(self):
        # print "Deleting test data"
        common.create_dbs(self._params, True)


    def test_staleness(self):
        # print "Querying map view"
        (map_resp, map_view_result) = common.query(self._params, "mapview1")
        map_etag = map_resp.getheader("ETag")

        self.assertEqual(map_view_result["total_rows"], self._params["ndocs"],
                         "Query returned %d total_rows" % self._params["ndocs"])
        self.assertEqual(len(map_view_result["rows"]), self._params["ndocs"],
                         "Query returned %d rows" % self._params["ndocs"])

        common.test_keys_sorted(map_view_result)

        # To give the set view group some time to write and fsync the index header
        time.sleep(5)

        # print "Verifying set view group info"
        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2, 3], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        for i in [0, 1, 2, 3]:
            self.assertEqual(info["update_seqs"][str(i)], (self._params["ndocs"] / 4),
                             "right update seq for partition %d" % (i + 1))

        # print "Adding 1 new document to each partition"
        server = self._params["server"]
        db1 = server[self._params["setname"] + "/0"]
        db2 = server[self._params["setname"] + "/1"]
        db3 = server[self._params["setname"] + "/2"]
        db4 = server[self._params["setname"] + "/3"]
        new_doc1 = {"_id": "new_doc_1", "integer": -1, "string": "1"}
        new_doc2 = {"_id": "new_doc_2", "integer": -2, "string": "2"}
        new_doc3 = {"_id": "new_doc_3", "integer": -3, "string": "3"}
        new_doc4 = {"_id": "new_doc_4", "integer": -4, "string": "4"}

        db1.save(new_doc1)
        db2.save(new_doc2)
        db3.save(new_doc3)
        db4.save(new_doc4)

        # print "Querying map view with ?stale=ok"
        (map_resp2, map_view_result2) = common.query(self._params, "mapview1", {"stale": "ok"})
        map_etag2 = map_resp2.getheader("ETag")

        self.assertEqual(map_view_result2, map_view_result, "Same response as before with ?stale=ok")
        self.assertEqual(map_etag2, map_etag, "Same etag as before with ?stale=ok")

        # print "Verifying set view group info"
        info2 = common.get_set_view_info(self._params)
        self.assertEqual(info2, info, "Same set view group info after ?stale=ok query")

        # print "Querying map view with ?stale=update_after"
        (map_resp3, map_view_result3) = common.query(self._params, "mapview1", {"stale": "update_after"})
        map_etag3 = map_resp3.getheader("ETag")

        self.assertEqual(map_view_result3, map_view_result, "Same response as before with ?stale=update_after")
        self.assertEqual(map_etag3, map_etag, "Same etag as before with ?stale=updater_after")

        time.sleep(5)

        # print "Verifying set view group info"
        info = common.get_set_view_info(self._params)
        self.assertEqual(info["active_partitions"], [0, 1, 2, 3], "right active partitions list")
        self.assertEqual(info["passive_partitions"], [], "right passive partitions list")
        self.assertEqual(info["cleanup_partitions"], [], "right cleanup partitions list")
        for i in [0, 1, 2, 3]:
            self.assertEqual(info["update_seqs"][str(i)], ((self._params["ndocs"] / 4) + 1),
                             "right update seq for partition %d" % (i + 1))

        # print "Querying map view without stale option"
        (map_resp4, map_view_result4) = common.query(self._params, "mapview1")
        map_etag4 = map_resp4.getheader("ETag")

        self.assertEqual(map_view_result4["total_rows"], (self._params["ndocs"] + 4),
                         "Query returned %d total_rows" % (self._params["ndocs"] + 4))
        self.assertEqual(len(map_view_result4["rows"]), (self._params["ndocs"] + 4),
                         "Query returned %d rows" % (self._params["ndocs"] + 4))
        self.assertNotEqual(map_etag4, map_etag3, "New etag after ?stale=update_after")

        common.test_keys_sorted(map_view_result4)

        self.assertEqual(map_view_result4["rows"][0]["id"], new_doc4["_id"], "new_doc4 in view result")
        self.assertEqual(map_view_result4["rows"][1]["id"], new_doc3["_id"], "new_doc3 in view result")
        self.assertEqual(map_view_result4["rows"][2]["id"], new_doc2["_id"], "new_doc2 in view result")
        self.assertEqual(map_view_result4["rows"][3]["id"], new_doc1["_id"], "new_doc1 in view result")

